import sys
import re
import pysam
import shutil
import os 

from multiprocessing import Pool
from optparse import OptionParser
from collections import Counter
from contextlib import contextmanager

opts = OptionParser()
usage = "usage: %prog [options] [inputs] Script to process aligned .bam files to 1) filter based on prevalent cell barcodes;  2) split based on valid chromosomes; and 3) filter for unique fragments (at the barcode level)"
opts = OptionParser(usage=usage)

opts.add_option("--input", "-i", help="Name of the .bam file to parse")
opts.add_option("--name", "-n", help="Name of the set of .bam files to collate")
opts.add_option("--output", "-o", help="Path to the output directory for these")

opts.add_option("--mapq", default = 30, help="Minimum mapq for a read to be kept")
opts.add_option("--barcode-whitelist", default = "none", help="Specify a whitelist of barcodes rather than determining them")

opts.add_option("--barcode-tag", default = 'XB', help="Name of the first .bam file")
opts.add_option("--min-fragments", default = 100, help="Minimum number of fragments for barcode consideration")
opts.add_option("--bedtools-genome", help="Filepath to bedtools genome.")
opts.add_option("--min_chromosome_size", default = 1000000, help="Minimum chromosome size (in bp) to be considered for fragment overlap analysis.")
opts.add_option("--ncores", default = 4, help="Number of cores for parallel processing.")
opts.add_option("--knee-call", default = "", help="File path to knee calling script")

options, arguments = opts.parse_args()

bamname = options.input
name = options.name
out = options.output

minmapq = float(options.mapq)
barcode_whitelist = options.barcode_whitelist

barcodeTag = options.barcode_tag
minFrag = int(options.min_fragments)
bedtoolsGenomeFile = options.bedtools_genome
minChromosomeSize = int(options.min_chromosome_size)
cpu = int(options.ncores)
knee_call_R = options.knee_call

def intersection(lst1, lst2):
	lst3 = [value for value in lst1 if value in lst2]
	return lst3

def listDictToCounter(lst):
	dct = Counter()
	for d in lst:
		for k, v in d.items():
			dct[k] += v 
	return(dct)

def getBarcode(intags):
	'''
	Parse out the barcode per-read
	'''
	for tg in intags:
		if(barcodeTag == tg[0]):
			return(tg[1])
	return("NA")

#---------------------------------------------------------------
# Function for extracting barcodes corresponding to unique reads
#---------------------------------------------------------------
def getUniqueBarcodes(chrom):

	# Note that the de-duplication logic here is such that
	# we don't care about the specific read being counted
	# as a singleton since we are only counting them.
	
	# Use dictionary logic
	barcodes = dict()
	barcodes_all = dict()
	
	# Keep track of the base pair for indexing unique reads
	barcode_bp = ['NA']
	bp = 0

	bam = pysam.AlignmentFile(bamname,'rb')
	Itr = bam.fetch(str(chrom),multiple_iterators=True)
	
	for read in Itr:
		if(read.mapping_quality > minmapq):
			read_barcode = getBarcode(read.tags)
			barcodes_all[read_barcode] = barcodes_all.get(read_barcode, 0) + 1
	
			# New base pair -- no duplicates
			if(read.reference_start != bp):
				bp = read.reference_start
				barcode_bp = [read_barcode]
				barcodes[read_barcode] = barcodes.get(read_barcode, 0) + 1
	
			# Same base pair -- verify that it's not existing barcodes
			else:
				if( not read_barcode in barcode_bp):
					barcodes[read_barcode] = barcodes.get(read_barcode, 0) + 1
					barcode_bp.append(read_barcode)
	return(barcodes, barcodes_all)


#---------------------------------------------------------
# Function for writing unique reads for barcodes that pass
#---------------------------------------------------------
def writeUniquePassingReads(chrom):
	
	# the bc variable is in the global environment
	
	# Note that the de-duplication logic here is such that
	# we DO care about whether the specific frag being reported
	# is consistent to ensure proper pairing downstream. The way
	# that this is achieved is by keeping a dictionary of reads per base pair index
	# and only writing the one that sort returns as the first
	
	# Keep track of the base pair for indexing unique reads
	barcode_bp_dict = dict()
	bp = 0
	bp_bc_count = 0
	idx = chrs.index(chrom)
	file = fopen[idx]

	# Iterate through bam file
	bam = pysam.AlignmentFile(bamname,'rb')
	Itr = bam.fetch(str(chrom),multiple_iterators=True)
	for read in Itr:

		# only consider reads with sufficient mapping quality
		if(read.mapping_quality > minmapq):
			read_barcode = getBarcode(read.tags)

			# New base pair -- no duplicates; write out the dictionary and update
			if(read.reference_start != bp):
				
				# Write out old base pair if we have things to write
				if(len(barcode_bp_dict) > 0):
					for key, value in barcode_bp_dict.items():
						value.tags = value.tags + [("NS", bp_bc_count)]
						file.write(value)
				
				# Now update to the new base pair... in part by wiping the dictionary
				barcode_bp_dict = dict()
				bp = read.reference_start
				barcode_bp_dict[read_barcode] = read
				bp_bc_count = 1
				
				# Else: same base pair -- do one of two things
				# 1) if we've seen the barcode before, then keep only the first sorted value
				# 2) if we haven't seen the barcode before, then append it
			else:
				# Still at the same base pair; verify that we haven't seen this barcode before
				bp_bc_count += 1
				if(read_barcode in barcode_bp_dict.keys()):
					old_read = barcode_bp_dict.get(read_barcode)
					old_name = old_read.query_name
					new_name = read.query_name
					
					# Keep newer read if it's a lesser read than the old one
					if(new_name < old_name):
						barcode_bp_dict[read_barcode] = read
					# else keep the old read-- no code needed
					
				# New barcode-- append the read directly
				else:
					barcode_bp_dict[read_barcode] = read
	bam.close()
	return(chrom)

# Handle the chromosomes
chrlens = {}
with open(bedtoolsGenomeFile) as f:
	for line in f:
		tok = line.split("\t")
		chrlens[tok[0]] = tok[1].strip()

# Split into those that are short (mitochondria) and long (nuclear)
chrlenfail = {x : chrlens[x] for x in chrlens if int(chrlens[x]) < minChromosomeSize }
chrsOther = list(chrlenfail.keys())

chrlenpass = {x : chrlens[x] for x in chrlens if int(chrlens[x]) >= minChromosomeSize }
chrs = list(chrlenpass.keys())

bamchrfiles = [out + "/" + name + "." + chr + ".raw" +".bam" for chr in chrs]
bamchrrouter = open(out.replace("temp/filt_split", ".internal/samples") + "/" + name + ".chrbam.txt", "w") 
for v in bamchrfiles:
	bamchrrouter.write(v+"\n")
bamchrrouter.close() 

	
# Quantify the barcodes into a list of dictionaries
pool = Pool(processes=cpu)
unique_barcodes, all_barcodes = zip(*pool.map(getUniqueBarcodes, chrs))
pool.close()

unique_barcodes = listDictToCounter(unique_barcodes)
all_barcodes = listDictToCounter(all_barcodes)

# Quantify read numbers for short chromosomes ==> mitochondria
pool = Pool(processes=cpu)
unique_bc_short, all_bc_short = zip(*pool.map(getUniqueBarcodes, chrsOther))
pool.close()
unique_bc_short = listDictToCounter(unique_bc_short)
all_bc_short = listDictToCounter(all_bc_short)

# Determine how bead barcodes are nominated-- also shared this information in a final bapParams file
paramsFile = open(out.replace("temp/filt_split", "knee") + "/" + name + ".bapParams.csv", "w")

if(os.path.isfile(barcode_whitelist)):
	knownBarcodes = True
	preDeterminedBarcodes =[line.rstrip('\n') for line in open(barcode_whitelist)]
	paramsFile.write("bead_threshold,predetermined\n")
	
elif(minFrag == 0):
	knownBarcodes = False
	
	# do the knee call
	quants_file=out.replace("temp/filt_split", "knee") + "/" + name + ".barcodeQuantSimple.csv"
	quants_handler = open(quants_file, 'w')
	for (k,v) in unique_barcodes.items():
		if(k != "NA"):
			quants_handler.write(str(k) + "," + str(v) + "\n")
	quants_handler.close()
	
	# Call Rscript 
	R_call = "Rscript " + knee_call_R + " " + quants_file + " 1 " + "V2"
	print("Performing knee call for bead barcode fragments--")
	os.system(R_call)
	with open(quants_file + "_kneeValue.txt") as knee_open:
		minFrag = float(knee_open.readline().strip())
		callFrag = float(knee_open.readline().strip())
	paramsFile.write("bead_threshold_nosafety,"+str(callFrag)+"\n")
	paramsFile.write("bead_threshold,"+str(minFrag)+"\n")
else:
	knownBarcodes = False
	paramsFile.write("bead_threshold,"+str(minFrag)+"\n")
paramsFile.close()

# Flatten list and determine barcodes passing filter or whitelist
if(knownBarcodes):
	barcodes = {x : unique_barcodes[x] for x in unique_barcodes if x in preDeterminedBarcodes}
else:
	barcodes = {x : unique_barcodes[x] for x in unique_barcodes if unique_barcodes[x] >= minFrag and x != "NA"}
global bc
bc = list(barcodes.keys())

#-------
# Loop back through, filter for positive barcodes, split by chr
#-------

# Function to open lots of files
@contextmanager
def multi_file_manager(files, mode='rt'):
	"""
	Open multiple files and make sure they all get closed.
	"""
	temp = pysam.AlignmentFile(bamname, "rb")
	files = [pysam.AlignmentFile(file, "wb", template = temp) for file in files]
	temp.close()
	yield files
	for file in files:
		file.close()
	
# Final loop to write out passing reads
with multi_file_manager(bamchrfiles) as fopen:
	pool = Pool(processes=cpu)
	toy_out = pool.map(writeUniquePassingReads, chrs)
	pool.close()

# Write out barcode file
bcfile = open(out.replace("temp/filt_split", "final") + "/" + name + ".barcodequants.csv", "w") 
bcfile.write("Barcode,UniqueNuclear,TotalNuclear,TotalMito"+ "\n")
for k, v in barcodes.items():
	if(all_bc_short.get(k) == None):
		mito = 0
	else:
		mito = all_bc_short.get(k)
	bcfile.write(k +","+ str(v)+"," + str(all_barcodes.get(k)) + "," + str(mito) + "\n")
bcfile.close() 




import click
import os
import os.path
import sys
import shutil
import random
import string
import itertools
import time
import platform
from ruamel import yaml
from pkg_resources import get_distribution
from .bapHelp import *

def getBfiles(bedtools_genome, blacklist_file, reference_genome, script_dir, supported_genomes):

	'''
	Function that isn't actually a bapProject specific function.
	Used to collate the built-in genomes with the possibility that the
	user specified another genome.
	'''
	
	# Handle bedtools
	if(bedtools_genome == "" and reference_genome == ""):
		sys.exit("ERROR: bap needs either the bedtools genome or a correctly specified reference genome to get peaks from summit files; QUITTING")
	elif any(reference_genome == s for s in supported_genomes):
		bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + reference_genome + ".sizes"
	else:
		if(os.path.isfile(bedtools_genome)):
			bedtoolsGenomeFile = bedtools_genome
		else: 
			sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
	
	# Handle blacklist	
	if(blacklist_file == "" and reference_genome == ""):
		sys.exit("ERROR: bap needs either a blacklist bed file or a correctly specified reference genome to get peaks from summit files; QUITTING")
	elif any(reference_genome == s for s in supported_genomes):
		blacklistFile = script_dir + "/anno/blacklist/" + reference_genome + ".full.blacklist.bed"
	else:
		if(os.path.isfile(bedtools_genome)):
			blacklistFile = bedtools_genome
		else: 
			sys.exit("Could not find the blacklist file: %s" % bedtools_genome)
			
	return(bedtoolsGenomeFile, blacklistFile)

def mitoChr(reference_genome, mito_chromosome):

	if(mito_chromosome != "default"):
		return(mito_chromosome)
	else:
		if(reference_genome in ["hg19", "mm10", "hg38", "mm10"]):
			return("chrM")
		elif(reference_genome in ["GRCh37", "GRCh38", "GRCm37", "GRCm38"]):
			return("MT")
		elif(reference_genome == "hg19_mm10_c"):
			return("humanM")
		else:
			return("hg19_chrM")

class bap2Project():
	def __init__(self, script_dir, supported_genomes, mode, input, output, name, ncores, reference_genome,
		cluster, jobs, peak_file, barcode_prior,
		minimum_barcode_fragments, barcode_whitelist,
		minimum_jaccard_index, nc_threshold, regularize_threshold, one_to_one, barcoded_tn5, keep_read_names,
		extract_mito, s_temp_files, snakemake_stdout, mapq, max_insert, all_pairs,
		bedtools_genome, blacklist_file, tss_file, mito_chromosome,
		r_path, bedtools_path, samtools_path, bgzip_path, tabix_path, snakemake_path,
		drop_tag, bead_tag, speciesMix):
		
		#----------------------------------
		# Assign straightforward attributes
		#----------------------------------
		self.bap_version = get_distribution('bap-atac').version
		self.script_dir = script_dir
		self.mode = mode
		self.output = output
		self.cluster = cluster
		self.jobs = jobs
		self.mapq = mapq
		self.ncores = ncores

		
		if(minimum_jaccard_index > 1):
			sys.exit("Cannot specify jaccard index > 1; user specified : %s" % str(minimum_jaccard_index))
		
		if(minimum_jaccard_index < 0):
			sys.exit("Cannot specify a nevative jaccard index; user specified : %s" % str(minimum_jaccard_index))
		
		
		# Handle potential whitelist
		if(barcode_whitelist == ""):
			barcode_whitelist = "none"
		else:
			if(os.path.isfile(barcode_whitelist)):
				pass
			else: 
				sys.exit("Could not find the bead whitelist file: %s" % barcode_whitelist)
				
		if(barcode_prior == ""):
			barcode_prior = "none"
		else:
			if(os.path.isfile(barcode_prior)):
				pass
			else: 
				sys.exit("Could not find the barcode prior file: %s" % barcode_prior)
		self.barcode_prior = barcode_prior
		self.barcode_whitelist = barcode_whitelist
		self.nc_threshold = nc_threshold
		self.regularize_threshold = regularize_threshold

		self.minimum_barcode_fragments = minimum_barcode_fragments
		self.minimum_jaccard_index = minimum_jaccard_index
		self.one_to_one = one_to_one
		self.barcoded_tn5 = barcoded_tn5
		self.keep_read_names = keep_read_names
		self.extract_mito = extract_mito
		self.drop_tag = drop_tag
		self.bead_tag = bead_tag
		self.max_insert = max_insert
		
		# Figure out operating system just for funzies; not presently used
		self.os = "linux"
		if(platform.platform()[0:5]=="Darwi"):
			self.os = "mac"
		
		if(mode == "bam"):
			self.bamfile = input
			self.bwa = "NA"
			self.bwa_index = "bwa_index"
			self.name = name
			
			if(name == "default"):
				filename, file_extension = os.path.splitext(self.bamfile)
				self.name = os.path.basename(filename)
		
		if(all_pairs):
			self.proper_pair = "false"
		else:
			self.proper_pair = "true"
		
		#------------------------------------------
		# Verify R and all of its packages are here
		#------------------------------------------
		R = get_software_path('R', r_path)
		#check_R_packages(['Rsamtools', 'GenomicAlignments', 'GenomicRanges', 'dplyr'], R)
		self.R = R
		
		bedtools = get_software_path('bedtools', bedtools_path)
		self.bedtools = bedtools
		samtools = get_software_path('samtools', samtools_path)
		self.samtools = samtools
		bgzip = get_software_path('bgzip', bgzip_path)
		self.bgzip = bgzip
		tabix = get_software_path('tabix', tabix_path)
		self.tabix = tabix
		snakemake = get_software_path('snakemake', snakemake_path)
		self.snakemake = snakemake
		#------------------------
		# Handle reference genome
		#------------------------
		self.reference_genome = reference_genome
		if any(self.reference_genome == s for s in supported_genomes):
			click.echo(gettime() + "Found designated reference genome: %s" % self.reference_genome)
			
			self.tssFile = script_dir + "/anno/TSS/" + self.reference_genome + ".refGene.TSS.bed"
			self.blacklistFile = script_dir + "/anno/blacklist/" + self.reference_genome + ".full.blacklist.bed"
			self.bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + self.reference_genome + ".sizes"

		else: 
			click.echo(gettime() + "Could not identify this reference genome: %s" % self.reference_genome)
			click.echo(gettime() + "Attempting to infer necessary input files from user specification.")
			necessary = [bedtools_genome, blacklist_file, tss_file, macs2_genome_size, bs_genome]
			if '' in necessary:
				if reference_genome == '':
					sys.exit("ERROR: specify valid reference genome with --reference-genome flag; QUITTING")
				else:
					sys.exit("ERROR: non-supported reference genome specified so these five must be validly specified: --bedtools-genome, --blacklist-file, --tss-file; QUITTING")
		
		if(reference_genome in ["hg19-mm10", "hg19_mm10_c"]):
			self.speciesMix = "yes"
		else:
			self.speciesMix = "none"
		
		#------------------------------		
		# Make sure all files are valid
		#------------------------------	
		if(bedtools_genome != ""):
			if(os.path.isfile(bedtools_genome)):
				self.bedtoolsGenomeFile = bedtools_genome
			else: 
				sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
				
		if(blacklist_file != ""):
			if(os.path.isfile(blacklist_file)):
				self.blacklistFile = blacklist_file
			else: 
				sys.exit("Could not find the blacklist bed file: %s" % blacklist_file)
		
		if(tss_file != ""):	
			if(os.path.isfile(tss_file)):
				self.tssFile = tss_file
			else: 
				sys.exit("Could not find the transcription start sites file: %s" % tss_file)
		
		if(peak_file != ""):	
			if(os.path.isfile(peak_file)):
				self.peakFile = peak_file
			else: 
				sys.exit("Could not find the transcription start sites file: %s" % peak_file)
		else:
			self.peakFile = ""
		
		self.mitochr = mitoChr(reference_genome, mito_chromosome)
		
	#--------------------------------------------------------------------------------
	# Define a method to dump the object as a .yaml/dictionary for use in other files
	#--------------------------------------------------------------------------------
	def __iter__(self):
		
		yield 'bap_version', self.bap_version
		yield 'script_dir', self.script_dir
		yield 'mode', self.mode
		yield 'output', self.output
		yield 'bamfile', self.bamfile
		yield 'name', self.name
		yield 'ncores', self.ncores
		
		yield 'cluster', self.cluster
		yield 'jobs', self.jobs
		yield 'peakFile', self.peakFile
		yield 'barcode_prior', self.barcode_prior
		
		yield 'minimum_barcode_fragments', self.minimum_barcode_fragments
		yield 'barcode_whitelist', self.barcode_whitelist

		yield 'minimum_jaccard_index', self.minimum_jaccard_index
		yield 'nc_threshold', self.nc_threshold
		yield 'regularize_threshold', self.regularize_threshold
		yield 'one_to_one', self.one_to_one
		yield 'barcoded_tn5', self.barcoded_tn5
		yield 'keep_read_names', self.keep_read_names
		yield 'proper_pair', self.proper_pair
		
		yield 'extract_mito', self.extract_mito
		yield 'mapq', self.mapq
		
		yield 'tssFile', self.tssFile
		yield 'blacklistFile', self.blacklistFile
		yield 'bedtoolsGenomeFile', self.bedtoolsGenomeFile
		yield 'mitochr', self.mitochr
		
		yield 'R', self.R
		yield 'samtools', self.samtools
		yield 'bedtools', self.bedtools
		yield 'bgzip', self.bgzip
		yield 'tabix', self.tabix
		yield 'snakemake', self.snakemake
		
		yield 'drop_tag', self.drop_tag
		yield 'bead_tag', self.bead_tag
		yield 'max_insert', self.max_insert
		yield 'speciesMix', self.speciesMix
	


#!/usr/bin/env python

import sys
import re
import pysam
from optparse import OptionParser

opts = OptionParser()
usage = "usage: %prog [options] [inputs] Software to process aligned bam files and annotate with a drop barcode with extracted mitochondria."
opts = OptionParser(usage=usage)
opts.add_option("--input", help="Filename of the existing .bam file")
opts.add_option("--output", help="Filename of new .bam file to be generated")
opts.add_option("--bead-barcode", help="Read tag of the bead barcode")
opts.add_option("--drop-barcode", help="Read tag of the drop barcode (to be added)")
opts.add_option("--dict-file", help="Filepath of the dictionary to convert bead-barcode to drop-barcode")
opts.add_option("--mitochr", help="Character string for the mitochondria")


options, arguments = opts.parse_args()

bamname = options.input
outname = options.output
bb = options.bead_barcode
db = options.drop_barcode
dictfile = options.dict_file
mitochr = options.mitochr

def getBarcode(intags):
	'''
	Parse out the bead barcode per-read
	'''
	for tg in intags:
		if(bb == tg[0]):
			return(tg[1])
	return("NA")

# Handle a dictionary of bead-barcode : drop-barcode pairs
d = {}
with open(dictfile) as f:
  for line in f:
    tok = line.split()
    d[tok[0]] = tok[1].strip()

bam = pysam.AlignmentFile(bamname, "rb")
out = pysam.AlignmentFile(outname, "wb", template = bam)

# Loop over bam and selectively write to bam de-duplicating droplet barcodes 
barcode_bp = ['NA']
bp = 0

try:
	for read in bam.fetch(mitochr):
		bead_bc = getBarcode(read.tags)
		drop_bc = d.get(bead_bc, "NA")
	
		# Handle droplet barcodes that we want to consider writing out
		if(drop_bc != "NA"):
	
			# New base pair -- no duplicates
			if(read.reference_start != bp):
				bp = read.reference_start
				barcode_bp = [drop_bc]
				read.tags = read.tags + [(db, drop_bc)]
				out.write(read)
				
			else:
				if( not drop_bc in barcode_bp):
					barcode_bp.append(drop_bc)
					read.tags = read.tags + [(db, drop_bc)]
					out.write(read)

except OSError: # Truncated bam file from previous iteration handle
	print('Finished parsing bam')
	
bam.close()
out.close()
pysam.index(outname)




import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time
import pysam

from pkg_resources import get_distribution
from subprocess import call, check_call
from .bapHelp import *
from .bap2ProjectClass import *
from ruamel import yaml
from ruamel.yaml.scalarstring import SingleQuotedScalarString as sqs

@click.command()
@click.version_option()

@click.argument('mode', type=click.Choice(['bam',  'check', 'support']))

@click.option('--input', '-i', help='Input for bap2-- a .bam file with an index.')
@click.option('--output', '-o', default="bap_out", help='Output directory for analysis; this is where everything is housed.')
@click.option('--name', '-n', default="default", help='Name for all of the output files (default: uses the .bam prefix)')

@click.option('--ncores', '-c', default = "detect", help='Number of cores to run the main job in parallel.')
@click.option('--reference-genome', '-r', default = "hg19", help='Specifcy supported reference genome. Check `bap2 support` for a list')

@click.option('--cluster', default = "",  help='Message to send to Snakemake to execute jobs on cluster interface; see Snakemake documentation.')
@click.option('--jobs', default = "0",  help='Max number of jobs to be running concurrently on the cluster interface; see Snakemake documentation.')

@click.option('--peak-file', '-pf', default = "", help='If supplied, compute FRIP (in QC stats) and generate Summarized Experiment (not required)')
@click.option('--barcode-prior', '-bp', default = "", help='If supplied, a two-column file that imposes barcode merges to occur within a prior annotation (e.g. donor or hash).')

@click.option('--minimum-barcode-fragments', '-bf', default = 0, help='Minimum number of fragments to be thresholded for doublet merging; by default, determines a threshold via knee-calling')
@click.option('--barcode-whitelist', '-w', default = "", help='File path of a whitelist of bead barcodes (one per line) to be used in lieu of a fixed threshold.')

@click.option('--minimum-jaccard-index', '-ji', default = 0.0, help='Minimum jaccard index for collapsing bead barcodes to cell barcodes')
@click.option('--nc-threshold', '-nc', default = 6, help='Number of barcodes that a paired-end read must be observed for the read to be filtered.')
@click.option('--regularize-threshold', '-rt', default = 4, help='Minimum number of inserts two barcodes must share to be counted per-chromosome. Anything less than this number will be shrunk to 0.')

@click.option('--one-to-one', '-oo', is_flag=True, help='Enforce that each bead barcode maps to one unique drop barcode (cancels the merging)')
@click.option('--barcoded-tn5',  is_flag=True, help='Process data knowing that the barcodes were generated with a barcoded Tn5; assumes that the last 6 characters of the barcode sequence correspond to the barcoded Tn5 sequence.')
@click.option('--keep-read-names',  is_flag=True, help='Retain the read namems that correspond to the final fragment assemebled in the final output.')

@click.option('--extract-mito', '-em', is_flag=True, help='Extract mitochondrial DNA and annotate with droplet barcodes.')
@click.option('--keep-temp-files', '-z', is_flag=True, help='Keep all intermediate files.')
@click.option('--snakemake-stdout', '-ss', is_flag=True, help='Write what would typically go in the snakemake log file into stdout.')

@click.option('--mapq', '-mq', default = 30, help='Minimum mapping quality to keep read for downstream analyses')
@click.option('--max-insert', '-mi', default = 2000, help='Max insert size to keep fragment for downstream analysis')
@click.option('--all-pairs', '-ap', is_flag=True, help='Include all read pairs when assembling fragments (not just proper pairs, which is the default)')

@click.option('--bedtools-genome', '-bg', default = "", help='Path to bedtools genome file; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--blacklist-file', '-bl', default = "", help='Path to bed file of blacklist; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--tss-file', '-ts', default = "", help='Path bed file of transcription start sites; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--mito-chromosome', '-mc', default = "default", help='Name of the mitochondrial chromosome; only necessary to specify if the reference genome is unknown but will overwrite default if necessary')

@click.option('--r-path', default = "", help='Path to R; by default, assumes that R is in PATH')
@click.option('--bedtools-path', default = "", help='Path to bedtools; by default, assumes that bedtools is in PATH')
@click.option('--samtools-path', default = "", help='Path to samtools; by default, assumes that samtools is in PATH')
@click.option('--bgzip-path', default = "", help='Path to bgzip; by default, assumes that bgzip is in PATH')
@click.option('--tabix-path', default = "", help='Path to tabix; by default, assumes that tabix is in PATH')
@click.option('--snakemake-path', default = "", help='Path to snakemake; by default, assumes that snakemake is in PATH')

@click.option('--drop-tag', '-dt', default = "DB", help='New tag in the .bam file(s) that will be the name of the droplet barcode.')
@click.option('--bead-tag', '-bt', default = "XB", help='Tag in the .bam file(s) that points to the bead barcode (should already exist).')


def main(mode, input, output, name, ncores, reference_genome,
	cluster, jobs, peak_file, barcode_prior,
	minimum_barcode_fragments, barcode_whitelist,
	minimum_jaccard_index, nc_threshold, regularize_threshold, one_to_one, barcoded_tn5, keep_read_names,
	extract_mito, keep_temp_files, snakemake_stdout, mapq, max_insert, all_pairs,
	bedtools_genome, blacklist_file, tss_file, mito_chromosome,
	r_path, bedtools_path, samtools_path, bgzip_path, tabix_path, snakemake_path,
	drop_tag, bead_tag):
	
	"""
	bap2: ~improved~ Bead-based scATAC-seq data Processing \n
	Caleb Lareau, clareau <at> broadinstitute <dot> org \n
	
	mode = ['bam', 'check', 'support']\n
	"""
	
	__version__ = get_distribution('bap-atac').version
	script_dir = os.path.dirname(os.path.realpath(__file__))

	click.echo(gettime() + "Starting bap2 pipeline v%s" % __version__)
	output = output.rstrip("/") # just for consistency
	
	# Determine which genomes are available
	rawsg = os.popen('ls ' + script_dir + "/anno/bedtools/*.sizes").read().strip().split("\n")
	supported_genomes = [x.replace(script_dir + "/anno/bedtools/chrom_", "").replace(".sizes", "") for x in rawsg]  
	
	# Determine number of cores in main job
	if(ncores == "detect"):
		ncores = str(available_cpu_count())
	else:
		ncores = str(ncores)
	
	# Parameterize optional snakemake configuration
	snakeclust = ""
	njobs = int(jobs)
	if(njobs > 0 and cluster != ""):
		snakeclust = " --jobs " + str(jobs) + " --cluster '" + cluster + "' "
		click.echo(gettime() + "Recognized flags to process jobs on a cluster.")
	

	if(mode == "support"):
		'''
		Show supported genomes and then bow out
		'''
		click.echo(gettime() + "List of built-in genomes supported in bap:")
		click.echo(gettime() + str(supported_genomes))
		sys.exit(gettime() + 'Specify one of these genomes or provide your own files (see documentation).')
	
	# Figure out if the specified reference genome is a species mix
	if(reference_genome in ["hg19-mm10", "hg19_mm10_c", "hg19-mm10_nochr"]):
		speciesMix = True
	else:
		speciesMix = False
	
	# Verify dependencies and set up an object to do all the dirty work
	p = bap2Project(script_dir, supported_genomes, mode, input, output, name, ncores, reference_genome,
		cluster, jobs, peak_file, barcode_prior,
		minimum_barcode_fragments, barcode_whitelist, 
		minimum_jaccard_index, nc_threshold, regularize_threshold, one_to_one, barcoded_tn5, keep_read_names,
		extract_mito, keep_temp_files, snakemake_stdout, mapq, max_insert, all_pairs,
		bedtools_genome, blacklist_file, tss_file, mito_chromosome,
		r_path, bedtools_path, samtools_path, bgzip_path, tabix_path, snakemake_path,
		drop_tag, bead_tag, speciesMix)
	
	# Make a counts table from user-supplied peaks and bam files
	if(mode == 'bam'):
	
		click.echo(gettime() + "Attempting to parse supplied .bam files for analysis.")
	
		# Make sure that the .bam file is valid
		if(not os.path.exists(input)):
			sys.exit("Cannot parse supplied .bam file in --input")
		if(not os.path.exists(input + ".bai")):
			sys.exit("Index supplied .bam before proceeding")
		
		# Verify that the supplied reference genome and the bam have overlapping chromosomes
		idxstats = pysam.idxstats(input)
		chrs_bam = [x.split("\t")[0] for x in idxstats.split("\n")]
		chrs_ref = []
		with open(p.bedtoolsGenomeFile) as f:
			for line in f:
				chrs_ref.append(line.split("\t")[0])
		
		# Quant the overlaps and exit if no overlap
		chrs_both = intersection(chrs_bam, chrs_ref)
		if(len(chrs_both) > 0):
			click.echo(gettime() + "Found " + str(len(chrs_both)) + " chromosomes for analysis (including mitochondria).")
		else:
			sys.exit("Found no overlapping chromosomes between bam and reference. Check reference genome specification with the --reference-genome flag.")
					
		# Make output folders
		of = output; logs = of + "/logs"; logs_bedpe = logs+"/frag"; fin = of + "/final"; mito = of + "/mito"; temp = of + "/temp"
		temp_filt_split = temp + "/filt_split"; temp_frag_overlap = temp + "/frag_overlap"; knee = of + "/knee"
		temp_drop_barcode = temp + "/drop_barcode"
		
		folders = [of, logs, logs_bedpe, fin, mito, temp,
			temp_filt_split, temp_frag_overlap,temp_drop_barcode, knee,
			of + "/.internal/parseltongue", of + "/.internal/samples"]
	
		mkfolderout = [make_folder(x) for x in folders]
		
		# Create internal README files 
		if not os.path.exists(of + "/.internal/README"):
			with open(of + "/.internal/README" , 'w') as outfile:
				outfile.write("This folder creates important (small) intermediate; don't modify it.\n\n")
		if not os.path.exists(of + "/.internal/parseltongue/README"):	
			with open(of + "/.internal/parseltongue/README" , 'w') as outfile:
				outfile.write("This folder creates intermediate output to be interpreted by Snakemake; don't modify it.\n\n")
		if not os.path.exists(of + "/.internal/samples/README"):
			with open(of + "/.internal" + "/samples" + "/README" , 'w') as outfile:
				outfile.write("This folder creates samples to be interpreted by Snakemake; don't modify it.\n\n")
		
		#-------------------------------------------------------
		# Step 1- Filter and split input .bam file by chromosome
		#-------------------------------------------------------
		click.echo(gettime() + "Splitting input bam files by chromosome for parallel processing.")
		click.echo(gettime() + "User specified "+ncores+" cores for parallel processing.")
		
		line1 = 'python ' +script_dir+'/bin/python/20_names_split_filt.py --input '+p.bamfile
		line2 = ' --name ' + p.name + ' --output ' + temp_filt_split + ' --barcode-tag ' 
		line3 = p.bead_tag + " --bedtools-reference-genome " + p.bedtoolsGenomeFile 
		line4 = " --mito-chr " +p.mitochr + " --ncores " + str(ncores) + " --mapq " + str(mapq)
		
		
		filt_split_cmd = line1 + line2 + line3 + line4 
		os.system(filt_split_cmd)

		#----------------------------------------
		# Step 2 - Process fragments by Snakemake
		#----------------------------------------
		click.echo(gettime() + "Processing bam file using a Snakemake workflow. This is the most computationally intensive step.")
		# Round trip the .yaml of user configuration
		y_s = of + "/.internal/parseltongue/bap.object.bam.yaml"
		with open(y_s, 'w') as yaml_file:
			yaml.dump(dict(p), yaml_file, default_flow_style=False, Dumper=yaml.RoundTripDumper)
		os.system("cp " + y_s +  " " + logs + "/" + p.name + ".parameters.txt")
		
		# Assemble some log files for the snake file
		snake_stats = logs + "/" + p.name + ".snakemake.stats"
		snake_log = logs + "/" + p.name + ".snakemake.log"
		
		# Handle edge cases of needing to write snakemake stdout to log
		if(snakemake_stdout):
			snake_log_preference = ""
		else:
			snake_log_preference = ' &>' + snake_log
		
		snakecmd_chr = p.snakemake+snakeclust+' --snakefile '+script_dir+'/bin/snake/Snakefile.bap2.chr --cores '+ncores+' --config cfp="' + y_s + '" --stats '+snake_stats+snake_log_preference
		os.system(snakecmd_chr)
		
		# Check to make sure snakemake Processing worked
		finalBamFile = p.output + "/final/" + p.name + ".bap.bam"
		if(os.path.exists(finalBamFile)):
			click.echo(gettime() + "Finished processing via Snakemake.")
		else:
			sys.exit(gettime() + "ERROR: Snakemake execution failed. Check " + snake_log + " file for more information. If blank, try allocating more memory.")


		#-----------------------------------
		# Step 3 - Process mitochondria
		#-----------------------------------
		if(extract_mito):
			click.echo(gettime() + "Creating a new .bam file for mitochondria.")
			
			dict_file = fin+"/"+p.name+".barcodeTranslate.tsv"
			
			line1 = 'python ' +script_dir+'/bin/python/17_processMito.py --input '+p.bamfile
			line2 = ' --output ' + mito + "/" + p.name + ".mito.bam" + " --mitochr " + p.mitochr 
			line3 = ' --bead-barcode ' + p.bead_tag + ' --drop-barcode ' + p.drop_tag + " --dict-file " + dict_file
			mito_cmd = line1 + line2 + line3
			
			os.system(mito_cmd)
		
		
		#-------------------------------------------------------
		# Final-- remove intermediate files if necessary
		#-------------------------------------------------------
		if keep_temp_files:
			click.echo(gettime() + "Temporary files not deleted since --keep-temp-files was specified.")
		else:
			byefolder = of
			shutil.rmtree(byefolder + "/.internal")
			shutil.rmtree(byefolder + "/temp")
			os.remove(fin + "/"+p.name+".basicQC.tsv")
			os.remove(knee + "/" + p.name + ".kneesPlotted.txt")

			if(not extract_mito):
				shutil.rmtree(byefolder + "/mito")
			
			click.echo(gettime() + "Intermediate files successfully removed.")

			
	# If we get to this stage in the processing, then everything worked
	if (mode == "check"):
		click.echo(gettime() + "Dependencies and user-reported file paths OK")
	
	click.echo(gettime() + "Complete.")
	



#!/usr/bin/env python

import sys
import re
import pysam
from optparse import OptionParser

opts = OptionParser()
usage = "usage: %prog [options] [inputs] Software to process aligned bam files and annotate with the number of occurences of a particular read across the experiment."
opts = OptionParser(usage=usage)
opts.add_option("--input", help="Filename of the existing .bam file")
opts.add_option("--output", help="Filename of new .bam file to be generated")
opts.add_option("--dict-file", help="Filepath of the dictionary to convert annotate with NC count")
opts.add_option("--nc-filt", help="Minimum permissible number of barcodes for the read to be detected without filtering it out")
opts.add_option("--bead-barcode", help="bead barcode tag in bam")

options, arguments = opts.parse_args()

bamname = options.input
outname = options.output
dictfile = options.dict_file
ncfilt = options.nc_filt
bb = options.bead_barcode

def get_barcode(intags):
	'''
	Parse out the bead barcode per-read
	'''
	for tg in intags:
		if(bb == tg[0]):
			return(tg[1])
	return("NA")

# Handle a dictionary of read-name : count pairs
d = {}
with open(dictfile) as f:
  for line in f:
    tok = line.split()
    d[tok[0]] = int(tok[1].strip())

bam = pysam.AlignmentFile(bamname, "rb")
out = pysam.AlignmentFile(outname, "wb", template = bam)

# Loop over bam and selectively write to bam de-duplicating droplet barcodes 
# Retain the same style as before 
bp = 0
bp_count = 0
barcode_blacklist_count = dict()

try:
	for read in bam:
		read_name = read.query_name
		read_count = d.get(read_name, 0)
		bead_bc = get_barcode(read.tags)
		
		# Check to verify that the read NC value is less than that which is permissible
		if(read_count <= int(ncfilt)):
			read.tags = read.tags + [("NC", read_count)]
			out.write(read)
		else:
			# Read is blacklisted; add a count to the barcode
			barcode_blacklist_count[bead_bc] = barcode_blacklist_count.get(bead_bc, 0) + 1
			
			
except OSError: # Truncated bam file from previous iteration handle
	print('Finished parsing bam')
	
bam.close()
out.close()

if(len(barcode_blacklist_count) > 0):
	with open(dictfile.replace("_ncRead.tsv", "_ncBarcode.tsv"), 'w') as file:
		for key, value in barcode_blacklist_count.items():
			file.write(key + "\t" + str(value) + "\n")




#!/usr/bin/env python

import sys
import re
import pysam
from optparse import OptionParser

opts = OptionParser()
usage = "usage: %prog [options] [inputs] Software to process aligned bam files and annotate with a drop barcode"
opts = OptionParser(usage=usage)
opts.add_option("--input", help="Filename of the existing .bam file")
opts.add_option("--output", help="Filename of new .bam file to be generated")
opts.add_option("--bead-barcode", help="Read tag of the bead barcode")
opts.add_option("--drop-barcode", help="Read tag of the drop barcode (to be added)")
opts.add_option("--dict-file", help="Filepath of the dictionary to convert bead-barcode to drop-barcode")

options, arguments = opts.parse_args()

bamname = options.input
outname = options.output
bb = options.bead_barcode
db = options.drop_barcode
dictfile = options.dict_file

def getBarcode(intags):
	'''
	Parse out the bead barcode per-read
	'''
	for tg in intags:
		if(bb == tg[0]):
			return(tg[1])
	return("NA")

# Handle a dictionary of bead-barcode : drop-barcode pairs
d = {}
with open(dictfile) as f:
  for line in f:
    tok = line.split()
    d[tok[0]] = tok[1].strip()

bam = pysam.AlignmentFile(bamname, "rb")
out = pysam.AlignmentFile(outname, "wb", template = bam)

# Loop over bam and selectively write to bam de-duplicating droplet barcodes 
# Retain the same style as before 
bp = 0
bp_count = 0
barcode_bp_dict = dict()

try:
	for read in bam:
		bead_bc = getBarcode(read.tags)
		drop_bc = d.get(bead_bc, "NA")
	
		# Handle droplet barcodes that we want to consider writing out
		if(drop_bc != "NA"):
			
			# Append the new tag to the read
			read.tags = read.tags + [(db, drop_bc)]
			
			# New base pair -- no duplicates; write out the dictionary and update
			if(read.reference_start != bp):
				
				# Write out old base pair if we have things to write
				if(len(barcode_bp_dict) > 0):
					for key, value in barcode_bp_dict.items():
						#value.tags = value.tags + [("NC", bp_count)]
						out.write(value)
				
				# Now update to the new base pair... in part by wiping the dictionary
				barcode_bp_dict = dict()
				bp = read.reference_start
				barcode_bp_dict[drop_bc] = read
				bp_count = 1
				
				# Else: same base pair -- do one of two things
				# 1) if we've seen the barcode before, then keep only the first sorted value
				# 2) if we haven't seen the barcode before, then append it
				
			else:
				# Still at the same base pair; verify that we haven't seen this barcode before
				bp_count += 1
				if(drop_bc in barcode_bp_dict.keys()):
					old_read = barcode_bp_dict.get(drop_bc)
					old_name = old_read.query_name
					new_name = read.query_name
					
					# Keep newer read if it's a lesser read than the old one
					if(new_name < old_name):
						barcode_bp_dict[drop_bc] = read
					# else keep the old read-- no code needed
					
				# New barcode-- append the read directly
				else:
					barcode_bp_dict[drop_bc] = read

except OSError: # Truncated bam file from previous iteration handle
	print('Finished parsing bam')
	
bam.close()
out.close()



import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time
import pysam

import csv
import re
from itertools import groupby
from .bapHelp import *
from pkg_resources import get_distribution

@click.command()
@click.version_option()

@click.option('--input', '-i', help='Input bam file.')
@click.option('--output', '-o', help='Output bam file.')

@click.option('--sep', '-s', default = "_", help='Separator for reannotation. Assume: {barcode}_{readname} (default delim = "_")')
@click.option('--tag', '-t', default="XB", help='Sam tag for barcode; by default, assume XB.\n\n')

def main(input, output, sep, tag):
	
	"""
	bap-reanno: Reannotate samples that were de-barcoded and aligned \n
	Caleb Lareau, clareau <at> broadinstitute <dot> org \n	
	"""
	
	__version__ = get_distribution('bap-atac').version
	script_dir = os.path.dirname(os.path.realpath(__file__))
	click.echo(gettime() + "Starting re-barcoding from bap pipeline v%s" % __version__)
	
	bam = pysam.AlignmentFile(input, "rb")
	out = pysam.AlignmentFile(output, "wb", template = bam)

	# Loop over bam and extract the sequence 
	for read in bam:
		name = read.query_name
		ss = name.split(sep)
		read.query_name = ss[1]
		read.tags = read.tags + [(tag, ss[0])]
		out.write(read)




import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time

import csv
import re
from itertools import groupby
from ..bapHelp import *
from pkg_resources import get_distribution
from subprocess import call, check_call

@click.command()
@click.version_option()

@click.argument('mode', type=click.Choice(['v1.0', 'v2.0', 'v2.1', 'v2.1-multi', '10X-v1']))

@click.option('--fastq1', '-a', help='Read 1 of the biological sequence; should contain the inline barcode in the sequence for use with BioRad parsing.')
@click.option('--fastq2', '-b', help='Read 2 of the biological sequence.')
@click.option('--fastqI', '-i', default = "", help='Fastq associated with the index read that is the cell barcode-- only needed for 10X parsing.')

@click.option('--output', '-o', default="debarcode", help='Output prefix for processed fastq files. By default, a simple string in the execution directory.\n\n')

@click.option('--ncores', '-c', default=2, help='Number of cores to be used in parallel de-barcoding')
@click.option('--nreads', '-nr', default=5000000, help='Number of reads to be processed in a chunk (both for an output file unit and in parallel processing)')
@click.option('--nmismatches', '-nm', default=1, help='Number of mismatches to be tolerated when doing a section of a barcode matching')
@click.option('--reverse-complement', '-rc', is_flag=True,  help='Perform the reverse complement of the barcodes - useful for 10X processing')


def main(mode, fastq1, fastq2, fastqi, output, ncores, nreads, nmismatches, reverse_complement):
	
	"""
	bap-barcode: De-barcode samples from BioRad bead single cell atac \n
	Caleb Lareau, clareau <at> broadinstitute <dot> org \n
	
	mode = ['v1.0', 'v2.0', 'v2.1', 'v2.1-multi', '10X-v1'] for bead design\n
	"""
	
	__version__ = get_distribution('bap').version
	script_dir = os.path.dirname(os.path.realpath(__file__))
	click.echo(gettime() + "Starting de-barcoding from bap pipeline v%s" % __version__)
	
	# Parse user settings
	core_call1 = " --fastq1 " + fastq1 + " --fastq2 " + fastq2 + " --ncores " + str(ncores) 
	core_call2 = " --nreads " + str(nreads) + " --nmismatches " + str(nmismatches) + " --output " + output
	core_call = core_call1 + core_call2
	
	# Handle mode to handle the configuration and make the right system call
	if(mode == "v1.0"):
		cmd = 'python '+script_dir+'/modes/biorad_v1.py '
		earlier = " --constant1 " + "TAGCCATCGCATTGC" + " --constant2 " + "TACCTCTGAGCTGAA"
		later = " --nextera " + "TCGTCGGCAGCGTC" + " --me " + "AGATGTGTATAAGAGACAG"
	elif(mode == "v2.0"):
		cmd = 'python '+script_dir+'/modes/biorad_v2.py '
		earlier = " --constant1 " + "TATGCATGAC" + " --constant2 " + "AGTCACTGAG"
		later = " --nextera " + "TGGTAGAGAGGGTG" + " --me " + "AGATGTGTATAAGAGACAG"
	elif(mode == "v2.1"):
		cmd = 'python '+script_dir+'/modes/biorad_v2.py '
		earlier = " --constant1 " + "TATGCATGAC" + " --constant2 " + "AGTCACTGAG"
		later = " --nextera " + "TCGTCGGCAGCGTC" + " --me " + "AGATGTGTATAAGAGACAG"
	elif(mode == "v2.1-multi"):
		cmd = 'python '+script_dir+'/modes/biorad_v2-multi.py '
		earlier = " --constant1 " + "TATGCATGAC" + " --constant2 " + "AGTCACTGAG"
		later = " --nextera " + "TCGTCGGCAGCGTC" + " --me " + "AGATGTGTATAAGAGACAG"
	elif(mode == "10X-v1"):
		if(reverse_complement):
			rc = " --reverse-complement 1 "
		else:
			rc = " --reverse-complement 0 "
		cmd = 'python '+script_dir+'/modes/tenX_v1.py '
		earlier = " --barcodesFile " + script_dir+'/modes/10Xdata/737K-cratac-v1.txt.gz'
		later = " --fastqI " + fastqi + rc
	else:
		sys.exit(gettime() + "User-supplied mode %s not found!" % mode)
	
	# Assemble the final call
	sys_call = cmd + earlier + later + core_call
	os.system(sys_call)
		



#!/usr/bin/env python

# Caleb Lareau, Broad Institute
# Implemented: 13 June 2018
# This program will demultiplex BioRad
# scATAC from v2.1 scheme

##### IMPORT MODULES #####
import os
import re
import regex
import sys
import gzip
from barcodeHelp import * # local python script

from optparse import OptionParser
from multiprocessing import Pool, freeze_support
from itertools import repeat

from Bio import SeqIO
from Bio.SeqIO.QualityIO import FastqGeneralIterator

from fuzzywuzzy import fuzz
from fuzzywuzzy import process
from fuzzysearch import find_near_matches

#### OPTIONS ####
opts = OptionParser()
usage = "usage: %prog [options] [inputs] Software to process raw .fastq reads and make data suitable for downstream processes"

opts.add_option("-a", "--fastq1", help="<Read1> Accepts fastq or fastq.gz")
opts.add_option("-b", "--fastq2", help="<Read2> Accepts fastq or fastq.gz")

opts.add_option("-n", "--nreads", default = 5000000, help="Number of reads in each split output file")
opts.add_option("-c", "--ncores", default = 4, help="Number of cores for parallel processing.")

opts.add_option("-j", "--constant1", default = "TAGCCATCGCATTGC", help="Barcode Constant 1")
opts.add_option("-k", "--constant2", default="TACCTCTGAGCTGAA", help="Barcode Constant 2")
opts.add_option("-l", "--nextera", default="TCGTCGGCAGCGTC", help="Nextera Adaptor Sequence")
opts.add_option("-m", "--me", default="AGATGTGTATAAGAGACAG", help="ME Sequence")

opts.add_option("-x", "--nmismatches", default=1, help="Number of mismatches")
opts.add_option("-o", "--output", help="Output sample convention")

options, arguments = opts.parse_args()

print(options)

# return usage information if no argvs given
if len(sys.argv)==1:
	os.system(sys.argv[0]+" --help")
	sys.exit()

##### INPUTS #####
a = options.fastq1
b = options.fastq2
outname = options.output
o = options.output

cpu = int(options.ncores)
n = int(options.nreads)

c1 = options.constant1
c2 = options.constant2
nxt = options.nextera
me = options.me
n_mismatch = int(options.nmismatches)

# Infer the length from the adaptors
c1_len = len(c1)
c2_len = len(c2)
nxt_len = len(nxt)
me_len = len(me)

# Parse input files
extension = a.split('.')[-1]
if extension == "fastq" or extension == "fq":
	sys.exist("Quitting... GZIP your .fastq files!")
elif extension == "gz":
	print("Found supplied .fastq.gz files")
else:
	sys.exit("ERROR! The input files (-a , -b) a *.fastq.gz")


# Define global variables
dumb = "N"*6 + "_" + "N"*6 + "_" + "N"*6
dumb2 = "N"*18

# Define barcodes
barcodes = ["AAAGAA","AACAGC","AACGTG","AAGCCA","AAGTAT","AATTGG","ACAAGG","ACCCAA","ACCTTC",
	"ACGGAC","ACTGCA","AGACCC","AGATGT","AGCACG","AGGTTA","AGTAAA","AGTCTG","ATACTT","ATAGCG",
	"ATATAC","ATCCGG","ATGAAG","ATTAGT","CAACCG","CAAGTC","CACCAC","CACTGT","CAGACT","CAGGAG","CATAGA",
	"CCACGC","CCGATG","CCGTAA","CCTCTA","CGAAAG","CGAGCA","CGCATA","CGGCGT","CGGTCC","CGTTAT","CTAGGT",
	"CTATTA","CTCAAT","CTGTGG","CTTACG","CTTGAA","GAAATA","GAAGGG","GACTCG","GAGCTT","GAGGCC","GAGTGA",
	"GATCAA","GCCAGA","GCCGTT","GCGAAT","GCGCGG","GCTCCC","GCTGAG","GCTTGT","GGACGA","GGATTG","GGCCAT",
	"GGGATC","GGTAGG","GGTGCT","GTACAG","GTCCTA","GTCGGC","GTGGTG","GTTAAC","GTTTCA","TAAGCT","TAATAG",
	"TACCGA","TAGAGG","TATTTC","TCAGTG","TCATCA","TCCAAG","TCGCCT","TCGGGA","TCTAGC","TGAATT","TGAGAC",
	"TGCGGT","TGCTAA","TGGCAG","TGTGTA","TGTTCG","TTAAGA","TTCGCA","TTCTTG","TTGCTC","TTGGAT","TTTGGG"
	]

#------------------------------

def extract_barcode_v1(sequence1):
	'''
	Function to extract barcodes
	'''

	# Parse out barcodes if we can ID the constants
	try:
		
		# use some approximate, yet generous, indices to facilitate faster matching
		c1_hit = find_near_matches(c1, sequence1[0:25],  max_l_dist=2) 
		c2_hit = find_near_matches(c2, sequence1[20:55], max_l_dist=2)
		nxt_hit = find_near_matches(nxt, sequence1[47:68],  max_l_dist=2)
		me_hit = find_near_matches(me, sequence1[55:95], max_l_dist=2)
		
		# Now grab the barcodes
		bc1, mm1 = prove_barcode(sequence1[c1_hit[0][0]-6:c1_hit[0][0]], barcodes, n_mismatch)
		bc2, mm2 = prove_barcode(sequence1[c1_hit[0][1]:c2_hit[0][0]+20], barcodes, n_mismatch)
		bc3, mm3 = prove_barcode(sequence1[c2_hit[0][1]+20:nxt_hit[0][0]+47], barcodes, n_mismatch)
		seq = sequence1[me_hit[0][1]+55:]
	
		return(bc1 + "_" + bc2 + "_" + bc3, seq, str(mm1)+","+str(mm2)+","+str(mm3))
	except:
		return(dumb, sequence1, "0,0,0")
	

def debarcode_v1(duo):
	"""
	Function that is called in parallel
	"""
	# Parse out inputs
	listRead1 = duo[0]; listRead2 = duo[1]
	
	# parameters to return
	fq1 = ""
	fq2 = ""
	mm_quant = ""

	nbc1 = 0
	nbc2 = 0
	nbc3 = 0

	npass = 0
	nfail = 0
	
	# Grab attributes
	title1 = listRead1[0]; sequence1 = listRead1[1]; quality1 = listRead1[2]
	title2 = listRead2[0]; sequence2 = listRead2[1]; quality2 = listRead2[2]
	
	# Return the barcode with underscores + the biological sequence learned	
	barcode, sequence1, mm = extract_barcode_v1(sequence1)
	quality1 = quality1[-1*len(sequence1):]
	
	three = barcode.split("_")
	barcode = "".join(three)
	
	if("NNNNNN" in three or len(sequence1) < 10):

		# Something went wrong
		nfail = nfail + 1
		if(barcode != dumb2):
			if("NNNNNN" == three[0]):
				nbc1 += 1
			if("NNNNNN" == three[1]):
				nbc2 += 1
			if("NNNNNN" == three[2]):
				nbc3 += 1

	else:
		npass = 1
		fq1 = formatRead("".join(three) + "_" + title1, sequence1, quality1)
		fq2 = formatRead("".join(three) + "_" + title2, sequence2, quality2)
		mm_quant = mm_quant + "".join(three) + "," + mm +"\n"
	return([[fq1, fq2], [nbc1, nbc2, nbc3, npass, nfail], [mm_quant]])


# Define variables to keep track of things that fail
nbc1 = 0
nbc2 = 0
nbc3 = 0

npass = 0
nfail = 0

with gzip.open(a, "rt") as f1:
	with gzip.open(b, "rt") as f2:
		
		# Establish iterators
		it1 = batch_iterator(FastqGeneralIterator(f1), n)
		it2 = batch_iterator(FastqGeneralIterator(f2), n)
		
		# iterate over batches of length n
		for i, batch1 in enumerate(it1):
			batch2 = it2.__next__()
			output = o +  "-c" + str(i+1).zfill(3)
			
			# parallel process the barcode processing and accounting of failures.
			pool = Pool(processes=cpu)
			pm = pool.map(debarcode_v1, zip(batch1, batch2))
			pool.close()
			
			# Aggregate output
			fqs = list(map(''.join, zip(*[item.pop(0) for item in pm])))
			counts = list(map(sum, zip(*[item.pop(0) for item in pm])))
			mm_values = list(map(''.join, zip(*[item.pop(0) for item in pm])))
			
			# Increment for QC
			nbc1 = nbc1 + counts[0]
			nbc2 = nbc2 + counts[1]
			nbc3 = nbc3 + counts[2]
			npass = npass + counts[3]
			nfail = nfail + counts[4]
			
			# Export one chunk in parallel
			filename1 = output +'_1.fastq.gz'
			filename2 = output +'_2.fastq.gz'
			#filenameMM = output +'_mismatches.csv.gz'
			
			pool = Pool(processes=2)
			toke = pool.starmap(chunk_writer_gzip, [(filename1, fqs[0]), (filename2, fqs[1])])
			pool.close()
			
with open(o + "-parse" + '.sumstats.log', 'w') as logfile:
	# give summary statistics
	logfile.write("\nParsing read pairs:\n" + a + "\n" + b + "\n")
	logfile.write("\n"+str(npass)+" reads parsed with barcodes ("+str(round(npass/(npass+nfail)*100, 2))+"% success)\n")
	logfile.write("Total reads that failed: "+str(nfail)+"\n\n")
	logfile.write("\nOf reads that could not be parsed that had valid, detectable constants:\n")
	logfile.write(str(nbc1) + " had a bad BC1 barcode sequence\n")
	logfile.write(str(nbc2) + " had a bad BC2 barcode sequence\n")
	logfile.write(str(nbc3) + " had a bad BC3 barcode sequence\n")
	
	


import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time
import pysam

from pkg_resources import get_distribution
from subprocess import call, check_call
from .bapHelp import *
from .bap1ProjectClass import *
from ruamel import yaml
from ruamel.yaml.scalarstring import SingleQuotedScalarString as sqs

@click.command()
@click.version_option()

@click.argument('mode', type=click.Choice(['bam', 'check', 'support']))

@click.option('--input', '-i', help='Input for bap; varies by which mode is specified; most certainly a .bam file with an index')
@click.option('--output', '-o', default="bap_out", help='Output directory for analysis; this is where everything is housed.')
@click.option('--name', '-n', default="default", help='Name for all of the output files (default: uses the .bam prefix)')

@click.option('--ncores', '-c', default = "detect", help='Number of cores to run the main job in parallel.')
@click.option('--reference-genome', '-r', default = "hg19", help='Support for built-in genome; choices are hg19, mm9, hg38, mm10, hg19_mm10_c and hg19-mm10 (species mix)')

@click.option('--cluster', default = "",  help='Message to send to Snakemake to execute jobs on cluster interface; see documentation.')
@click.option('--jobs', default = "0",  help='Max number of jobs to be running concurrently on the cluster interface.')
@click.option('--peak-file', '-pf', default = "", help='If supplied, compute FRIP (in QC stats) and generate Summarized Experiment')

@click.option('--minimum-barcode-fragments', '-bf', default = 0, help='Minimum number of fragments to be thresholded for doublet merging.')
@click.option('--minimum-cell-fragments', '-cf', default = 500, help='Minimum number of unique to be thresholded for final output.')
@click.option('--barcode-whitelist', '-wl', default = "", help='File path of a whitelist of bead barcodes (one per line) to be used in lieu of a fixed threshold.')

@click.option('--minimum-jaccard-index', '-ji', default = 0.0, help='Minimum jaccard index for collapsing bead barcodes to cell barcodes')
@click.option('--nc-threshold', '-nc', default = 6, help='Number of barcodes that a paired-end read must be observed for the read to be filtered.')
@click.option('--one-to-one', '-oo', is_flag=True, help='Enforce that each bead barcode maps to one unique drop barcode (making this merging useless)')
@click.option('--barcoded-tn5',  is_flag=True, help='Process data knowing that the barcodes were generated with a barcoded Tn5')

@click.option('--extract-mito', '-em', is_flag=True, help='Extract mitochondrial DNA too?.')
@click.option('--keep-temp-files', '-z', is_flag=True, help='Keep all intermediate files.')
@click.option('--mapq', '-mq', default = 30, help='Minimum mapping quality to keep read for downstream analyses')

@click.option('--bedtools-genome', '-bg', default = "", help='Path to bedtools genome file; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--blacklist-file', '-bl', default = "", help='Path to bed file of blacklist; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--tss-file', '-ts', default = "", help='Path bed file of transcription start sites; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--mito-chromosome', '-mc', default = "default", help='Name of the mitochondrial chromosome; only necessary to specify if the reference genome is unknown but will overwrite default if needbe')
@click.option('--r-path', default = "", help='Path to R; by default, assumes that R is in PATH')

@click.option('--drop-tag', '-dt', default = "DB", help='New tag in the .bam file(s) that will be the name of the drop barcode.')
@click.option('--bead-tag', '-bt', default = "XB", help='Tag in the .bam file(s) that point to the bead barcode; valid for bam mode.')


def main(mode, input, output, name, ncores, reference_genome,
	cluster, jobs, peak_file,
	minimum_barcode_fragments, minimum_cell_fragments, barcode_whitelist,
	minimum_jaccard_index, nc_threshold, one_to_one, barcoded_tn5,
	extract_mito, keep_temp_files, mapq, 
	bedtools_genome, blacklist_file, tss_file, mito_chromosome, r_path, 
	drop_tag, bead_tag):
	
	"""
	bap: Bead-based scATAC-seq data Processing \n
	Caleb Lareau, clareau <at> broadinstitute <dot> org \n
	
	mode = ['bam', 'check', 'support']\n
	"""
	
	__version__ = get_distribution('bap-atac').version
	script_dir = os.path.dirname(os.path.realpath(__file__))

	click.echo(gettime() + "Starting bap pipeline v%s" % __version__)
	click.echo(gettime() + "WARNING: this is bap1-- you should really use bap2 unless you are trying to reproduce old results!")
	click.echo(gettime() + "NOTE: The same arguments apply-- see `bap2 --help`")

	# Determine which genomes are available
	rawsg = os.popen('ls ' + script_dir + "/anno/bedtools/*.sizes").read().strip().split("\n")
	supported_genomes = [x.replace(script_dir + "/anno/bedtools/chrom_", "").replace(".sizes", "") for x in rawsg]  
	
	# Determine number of cores in main job
	if(ncores == "detect"):
		ncores = str(available_cpu_count())
	else:
		ncores = str(ncores)
	
	# Parameterize optional snakemake configuration
	snakeclust = ""
	njobs = int(jobs)
	if(njobs > 0 and cluster != ""):
		snakeclust = " --jobs " + str(jobs) + " --cluster '" + cluster + "' "
		click.echo(gettime() + "Recognized flags to process jobs on a cluster.")
	

	if(mode == "support"):
		'''
		Show supported genomes and then bow out
		'''
		click.echo(gettime() + "List of built-in genomes supported in bap:")
		click.echo(gettime() + str(supported_genomes))
		sys.exit(gettime() + 'Specify one of these genomes or provide your own files (see documentation).')
		
	# Verify dependencies and set up an object to do all the dirty work
	p = bap1Project(script_dir, supported_genomes, mode, input, output, name, ncores, reference_genome,
		cluster, jobs, peak_file,
		minimum_barcode_fragments, minimum_cell_fragments, barcode_whitelist,
		minimum_jaccard_index, nc_threshold, one_to_one, barcoded_tn5, 
		extract_mito, keep_temp_files, mapq, 
		bedtools_genome, blacklist_file, tss_file, mito_chromosome, r_path, 
		drop_tag, bead_tag)
	
	if(reference_genome in ["hg19-mm10", "hg19_mm10_c", "hg19-mm10_nochr"]):
		speciesMix = True
	else:
		speciesMix = False
	
	# Make a counts table from user-supplied peaks and bam files
	if(mode == 'bam'):
	
		click.echo(gettime() + "Attempting to parse supplied .bam files for analysis.")
	
		# Make sure that the .bam file is valid
		if(not os.path.exists(input)):
			sys.exit("Cannot parse supplied .bam file in --input")
		if(not os.path.exists(input + ".bai")):
			sys.exit("Index supplied .bam before proceeding")
		
		# Verify that the supplied reference genome and the bam have overlapping chromosomes
		idxstats = pysam.idxstats(input)
		chrs_bam = [x.split("\t")[0] for x in idxstats.split("\n")]
		chrs_ref = []
		with open(p.bedtoolsGenomeFile) as f:
			for line in f:
				chrs_ref.append(line.split("\t")[0])
		
		# Quant the overlaps and exit if no overlap
		chrs_both = intersection(chrs_bam, chrs_ref)
		if(len(chrs_both) > 0):
			click.echo(gettime() + "Found " + str(len(chrs_both)) + " chromosomes for analysis (including mitochondria).")
		else:
			sys.exit("Found no overlapping chromosomes between bam and reference. Check reference genome specification with the --reference-genome flag.")
					
		# Make output folders
		of = output; logs = of + "/logs"; fin = of + "/final"; mito = of + "/mito"; temp = of + "/temp"
		temp_filt_split = temp + "/filt_split"; temp_frag_overlap = temp + "/frag_overlap"; knee = of + "/knee"
		temp_drop_barcode = temp + "/drop_barcode"
		
		folders = [of, logs, fin, mito, temp,
			temp_filt_split, temp_frag_overlap,temp_drop_barcode, knee,
			of + "/.internal/parseltongue", of + "/.internal/samples"]
	
		mkfolderout = [make_folder(x) for x in folders]
		
		# Create internal README files 
		if not os.path.exists(of + "/.internal/README"):
			with open(of + "/.internal/README" , 'w') as outfile:
				outfile.write("This folder creates important (small) intermediate; don't modify it.\n\n")
		if not os.path.exists(of + "/.internal/parseltongue/README"):	
			with open(of + "/.internal/parseltongue/README" , 'w') as outfile:
				outfile.write("This folder creates intermediate output to be interpreted by Snakemake; don't modify it.\n\n")
		if not os.path.exists(of + "/.internal/samples/README"):
			with open(of + "/.internal" + "/samples" + "/README" , 'w') as outfile:
				outfile.write("This folder creates samples to be interpreted by Snakemake; don't modify it.\n\n")
		
		#-------------------------------------------------------
		# Step 1- Filter and split input .bam file by chromosome
		#-------------------------------------------------------		
		line1 = 'python ' +script_dir+'/bin/python/10_quantBarcode_Filt.py --input '+p.bamfile
		line2 = ' --name ' + p.name + ' --output ' + temp_filt_split + ' --barcode-tag ' 
		line3 = p.bead_tag + ' --min-fragments ' + str(p.minimum_barcode_fragments)
		line4 = " --bedtools-genome " +p.bedtoolsGenomeFile + " --ncores " + str(ncores) + " --mapq " + str(mapq)
		line5 = " --barcode-whitelist " + p.barcode_whitelist + " --knee-call " + script_dir+'/bin/R/10b_knee_execute.R'
			
		filt_split_cmd = line1 + line2 + line3 + line4 + line5
		print(filt_split_cmd)
		os.system(filt_split_cmd)
		
		# Verify that we got output and fail if not
		val = file_len(p.output + "/final/" + p.name + ".barcodequants.csv") -1
		click.echo(gettime() + "Found " + str(val) + " abundant barcodes passing filter.")
		
		if(val < 2):
			sys.exit(gettime() + "Found insufficient # of barcodes for analyis; check reference genome or lower --minimum-barcode-fragments")
		
		#----------------------------------------
		# Step 2 - Process fragments by Snakemake
		#----------------------------------------
		
		# Round trip the .yaml of user configuration
		y_s = of + "/.internal/parseltongue/bap.object.bam.yaml"
		with open(y_s, 'w') as yaml_file:
			yaml.dump(dict(p), yaml_file, default_flow_style=False, Dumper=yaml.RoundTripDumper)
			
		snakecmd_chr = 'snakemake'+snakeclust+' --snakefile '+script_dir+'/bin/snake/Snakefile.bap.chr --cores '+ncores+' --config cfp="' + y_s + '"'
		os.system(snakecmd_chr)
		
		#-------------------------------------------------
		# Step 3 - Handle aggregation of NC filtered reads
		#-------------------------------------------------
		nc_R = script_dir + "/bin/R/15_ncReadCleanup.R"
		barcodeQuantFile =  p.output + "/final/" + p.name + ".barcodequants.csv"
		
		# Irrespective, modify the barcode quantification sumstats
		nc_R_call = " ".join([p.R+"script", nc_R, p.output, str(keep_temp_files), barcodeQuantFile])
		os.system(nc_R_call)

		#-----------------------------------
		# Step 4 - QC stats / outside Snakemake since not-essential
		#-----------------------------------
		barcodeTranslateFile = p.output + "/final/" + p.name + ".barcodeTranslate.tsv"
		qcStats16File =  p.output + "/final/" + p.name + ".QCstats.csv"
		finalBamFile = p.output + "/final/" + p.name + ".bap.bam"
		qc_R = script_dir + "/bin/R/16_qualityControlReport_SE.R"
		
		# Determine user flags for proper QC stuff		
		if(p.peakFile == ""):
			peakFileGo = "none"
		else:
			peakFileGo = p.peakFile
		
		if(speciesMix):
			speciesMixGo = "yes"
		else:
			speciesMixGo = "no"
			
		if(one_to_one):
			oneToOneGo = "yes"
		else:
			oneToOneGo = "no"
			
		click.echo(gettime() + "generating QC report + summarized experiment file...")
		# Full system call to R script
		r_callQC = " ".join([p.R+"script", qc_R, finalBamFile, barcodeTranslateFile, barcodeQuantFile, p.tssFile, p.drop_tag, p.blacklistFile, peakFileGo, speciesMixGo, oneToOneGo])
		os.system(r_callQC)

		#-----------------------------------
		# Step 5 - Process mitochondria
		#-----------------------------------
		if(extract_mito):
			click.echo(gettime() + "Creating a new .bam file for mitochondria.")
			
			dict_file = fin+"/"+p.name+".barcodeTranslate.tsv"
			
			line1 = 'python ' +script_dir+'/bin/python/17_processMito.py --input '+p.bamfile
			line2 = ' --output ' + mito + "/" + p.name + ".mito.bam" + " --mitochr " + p.mitochr 
			line3 = ' --bead-barcode ' + p.bead_tag + ' --drop-barcode ' + p.drop_tag + " --dict-file " + dict_file
			mito_cmd = line1 + line2 + line3
			
			os.system(mito_cmd)
		
		#-----------------------------------
		# Step 6 - Make knee plots
		#-----------------------------------
		bapParamsFile =  p.output + "/knee/" + p.name + ".bapParams.csv"
		beadBarcodesFile = p.output + "/knee/" + p.name + ".barcodeQuantSimple.csv"
		implicatedBarcodeFile = p.output + "/final/" + p.name + ".implicatedBarcodes.csv.gz"
		
		click.echo(gettime() + "generating knee plots associated with beads.")
		kneePlot_R = script_dir + "/bin/R/19_makeKneePlots.R"
		r_callKneePlot = " ".join([p.R+"script", kneePlot_R, bapParamsFile, beadBarcodesFile, implicatedBarcodeFile])
		os.system(r_callKneePlot)
		
		#-------------------------------------------------------
		# Final-- remove intermediate files if necessary
		#-------------------------------------------------------
		if keep_temp_files:
			click.echo(gettime() + "Temporary files not deleted since --keep-temp-files was specified.")
		else:
			byefolder = of
			shutil.rmtree(byefolder + "/.internal")
			shutil.rmtree(byefolder + "/temp")

			if(not extract_mito):
				shutil.rmtree(byefolder + "/mito")
			
			click.echo(gettime() + "Intermediate files successfully removed.")
		
			
	# If we get to this stage in the processing, then everything worked
	if (mode == "check"):
		click.echo(gettime() + "Dependencies and user-reported file paths OK")
	
	click.echo(gettime() + "Complete.")
	



import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time
import pysam

from pkg_resources import get_distribution
from subprocess import call, check_call
from .bapHelp import *
from .bapFragProjectClass import *
from ruamel import yaml
from ruamel.yaml.scalarstring import SingleQuotedScalarString as sqs

@click.command()
@click.version_option()

@click.option('--input', '-i', help='Input for bap-frag; varies by which mode is specified; for `bam`, .bam file with an index.')
@click.option('--output', '-o', default="bap_out", help='Output directory for analysis; this is where everything is housed.')
@click.option('--name', '-n', default="default", help='Name for all of the output files (default: uses the .bam prefix)')

@click.option('--ncores', '-c', default = "detect", help='Number of cores to run the main job in parallel.')
@click.option('--reference-genome', '-r', default = "hg19", help='Specifcy supported reference genome. Check `bap2 support` for a list')

@click.option('--cluster', default = "",  help='Message to send to Snakemake to execute jobs on cluster interface; see documentation.')
@click.option('--jobs', default = "0",  help='Max number of jobs to be running concurrently on the cluster interface.')

@click.option('--keep-temp-files', '-z', is_flag=True, help='Keep all intermediate files.')
@click.option('--mapq', '-mq', default = 30, help='Minimum mapping quality to keep read for downstream analyses')

@click.option('--bedtools-genome', '-bg', default = "", help='Path to bedtools genome file; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--blacklist-file', '-bl', default = "", help='Path to bed file of blacklist; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--tss-file', '-ts', default = "", help='Path bed file of transcription start sites; overrides default if --reference-genome flag is set and is necessary for non-supported genomes.')
@click.option('--mito-chromosome', '-mc', default = "default", help='Name of the mitochondrial chromosome; only necessary to specify if the reference genome is unknown but will overwrite default if necessary')

@click.option('--r-path', default = "", help='Path to R; by default, assumes that R is in PATH')
@click.option('--bedtools-path', default = "", help='Path to bedtools; by default, assumes that bedtools is in PATH')
@click.option('--samtools-path', default = "", help='Path to samtools; by default, assumes that samtools is in PATH')
@click.option('--bgzip-path', default = "", help='Path to bgzip; by default, assumes that bgzip is in PATH')
@click.option('--tabix-path', default = "", help='Path to samtools; by default, assumes that tabix is in PATH')


def main(input, output, name, ncores, reference_genome,
	cluster, jobs,
	keep_temp_files, mapq,
	bedtools_genome, blacklist_file, tss_file, mito_chromosome,
	r_path, bedtools_path, samtools_path, bgzip_path, tabix_path):
	
	"""
	bap-frag: Extract fragments from pre-defined cells \n
	Caleb Lareau, clareau <at> broadinstitute <dot> org \n
	
	"""
	
	__version__ = get_distribution('bap-atac').version
	script_dir = os.path.dirname(os.path.realpath(__file__))

	click.echo(gettime() + "Starting bap-frag pipeline v%s" % __version__)
	
	# Determine which genomes are available
	rawsg = os.popen('ls ' + script_dir + "/anno/bedtools/*.sizes").read().strip().split("\n")
	supported_genomes = [x.replace(script_dir + "/anno/bedtools/chrom_", "").replace(".sizes", "") for x in rawsg]  
	
	# Determine number of cores in main job
	if(ncores == "detect"):
		ncores = str(available_cpu_count())
	else:
		ncores = str(ncores)
	
	# Parameterize optional snakemake configuration
	snakeclust = ""
	njobs = int(jobs)
	if(njobs > 0 and cluster != ""):
		snakeclust = " --jobs " + str(jobs) + " --cluster '" + cluster + "' "
		click.echo(gettime() + "Recognized flags to process jobs on a cluster.")

	# Figure out if the specified reference genome is a species mix
	if(reference_genome in ["hg19-mm10", "hg19_mm10_c"]):
		speciesMix = True
	else:
		speciesMix = False
	
	# Verify dependencies and set up an object to do all the dirty work
	p = bapBulkFragProject(script_dir, supported_genomes, input, output, name, ncores, reference_genome,
		cluster, jobs,
		keep_temp_files, mapq, 
		bedtools_genome, blacklist_file, tss_file, mito_chromosome,
		r_path, bedtools_path, samtools_path, bgzip_path, tabix_path)

	click.echo(gettime() + "Attempting to parse supplied .bam files for analysis.")

	# Make sure that the .bam file is valid
	if(not os.path.exists(input)):
		sys.exit("Cannot parse supplied .bam file in --input")
	if(not os.path.exists(input + ".bai")):
		sys.exit("Index supplied .bam before proceeding")
	
	# Verify that the supplied reference genome and the bam have overlapping chromosomes
	idxstats = pysam.idxstats(input)
	chrs_bam = [x.split("\t")[0] for x in idxstats.split("\n")]
	chrs_ref = []
	with open(p.bedtoolsGenomeFile) as f:
		for line in f:
			chrs_ref.append(line.split("\t")[0])
	
	# Quant the overlaps and exit if no overlap
	chrs_both = intersection(chrs_bam, chrs_ref)
	if(len(chrs_both) > 0):
		click.echo(gettime() + "Found " + str(len(chrs_both)) + " chromosomes for analysis (including mitochondria).")
	else:
		sys.exit("Found no overlapping chromosomes between bam and reference. Check reference genome specification with the --reference-genome flag.")
				
	# Make output folders
	of = output; logs = of + "/logs"; logs_bedpe = logs+"/frag"; fin = of + "/final"; temp = of + "/temp"
	temp_filt_split = temp + "/filt_split"
		
	folders = [of, logs, logs_bedpe, fin, temp,
		temp_filt_split, 
		of + "/.internal/parseltongue", of + "/.internal/samples"]

	mkfolderout = [make_folder(x) for x in folders]
	
	# Create internal README files 
	if not os.path.exists(of + "/.internal/README"):
		with open(of + "/.internal/README" , 'w') as outfile:
			outfile.write("This folder creates important (small) intermediate; don't modify it.\n\n")
	if not os.path.exists(of + "/.internal/parseltongue/README"):	
		with open(of + "/.internal/parseltongue/README" , 'w') as outfile:
			outfile.write("This folder creates intermediate output to be interpreted by Snakemake; don't modify it.\n\n")
	if not os.path.exists(of + "/.internal/samples/README"):
		with open(of + "/.internal" + "/samples" + "/README" , 'w') as outfile:
			outfile.write("This folder creates samples to be interpreted by Snakemake; don't modify it.\n\n")
	
	#-------------------------------------------------------
	# Step 1- Filter and split input .bam file by chromosome
	#-------------------------------------------------------
	click.echo(gettime() + "Splitting input bam files by chromosome for parallel processing.")
	click.echo(gettime() + "User specified "+ncores+" cores for parallel processing.")
	
	line1 = 'python ' +script_dir+'/bin/python/20a_nonames_split_filt.py --input '+p.bamfile
	line2 = ' --name ' + p.name + ' --output ' + temp_filt_split 
	line3 = " --bedtools-reference-genome " + p.bedtoolsGenomeFile 
	line4 = " --mito-chr " +p.mitochr + " --ncores " + str(ncores) + " --mapq " + str(mapq)
	
	filt_split_cmd = line1 + line2 + line3 + line4 
	os.system(filt_split_cmd)
	
	#----------------------------------------
	# Step 2 - Process fragments by Snakemake
	#----------------------------------------
	click.echo(gettime() + "Processing per-chromosome fragments in parallel. This is the most computationally intensive step.")
	# Round trip the .yaml of user configuration
	y_s = of + "/.internal/parseltongue/bap.object.bam.yaml"
	with open(y_s, 'w') as yaml_file:
		yaml.dump(dict(p), yaml_file, default_flow_style=False, Dumper=yaml.RoundTripDumper)
	os.system("cp " + y_s +  " " + logs + "/" + p.name + ".parameters.txt")
	
	# Assemble some log files for the snake file
	snake_stats = logs + "/" + p.name + ".snakemake.stats"
	snake_log = logs + "/" + p.name + ".snakemake.log"
	snakecmd_chr = 'snakemake'+snakeclust+' --snakefile '+script_dir+'/bin/snake/Snakefile.bap_bulk_frags --cores '+ncores+' --config cfp="' + y_s + '" --stats '+snake_stats+' &>' + snake_log
	os.system(snakecmd_chr)
		
	#-------------------------------------------------------
	# Final-- remove intermediate files if necessary
	#-------------------------------------------------------
	if keep_temp_files:
		click.echo(gettime() + "Temporary files not deleted since --keep-temp-files was specified.")
	else:
		byefolder = of
		shutil.rmtree(byefolder + "/.internal")
		shutil.rmtree(byefolder + "/temp")
		
		click.echo(gettime() + "Intermediate files successfully removed.")

	click.echo(gettime() + "Complete.")
	



import itertools
import time
import shutil
import re
import os
import sys
import csv
import gzip
from ruamel import yaml
from functools import partial

def string_hamming_distance(str1, str2):
	'''
	Fast hamming distance over 2 strings known to be of same length.
	In information theory, the Hamming distance between two strings of equal 
	length is the number of positions at which the corresponding symbols 
	are different.
	eg "karolin" and "kathrin" is 3.
	'''
	return sum(itertools.imap(operator.ne, str1, str2))

def intersection(lst1, lst2):
	lst3 = [value for value in lst1 if value in lst2]
	return lst3

def make_folder(folder):
	"""
	Function to only make a given folder if it does not already exist
	"""
	if not os.path.exists(folder):
		os.makedirs(folder)

def file_len(fname):
	"""
	Function to count number of lines in a file
	"""
	with open(fname) as f:
		for i, l in enumerate(f):
			pass
	return i + 1

def verify_file(filename):
	"""
	Ensure that file can both be read and exists
	"""
	try:
		extension = filename.split('.')[-1]
		if extension == "gz":
			fp = gzip.open(filename, 'rt')
		else:
			fp = open(filename)
		one = fp.readline()
		fp.close()
	except IOError as err:
		sys.exit(gettime() + "Error reading the file {0}: {1}".format(filename, err))
	return(filename)


def get_software_path(tool, abs_path):
	'''
	Function takes a tool name and a possible absolute path
	specified by the user input and returns the absolute path
	to the tool
	'''
	tool_path = shutil.which(tool)
	if(str(tool_path) == "None"):
		if(abs_path == ""):
			sys.exit("ERROR: cannot find "+tool+" in environment; add it to user PATH environment or specify executable using a flag.")
	if(abs_path != ""):
		if(os.path.isfile(abs_file)):
			tool_path = abs_path
	return(tool_path)

def rev_comp(seq):
	'''
	Fast Reverse Compliment
	'''  
	tbl = {'A':'T', 'T':'A', 'C':'G', 'G':'C', 'N':'N'}
	return ''.join(tbl[s] for s in seq[::-1])

def gettime(): 
	'''
	Matches `date` in unix
	'''
	return(time.strftime("%a ") + time.strftime("%b ") + time.strftime("%d ") + time.strftime("%X ") + 
		time.strftime("%Z ") + time.strftime("%Y")+ ": ")
		

def findIdx(list1, list2):
	'''
	Return the indices of list1 in list2
	'''
	return [i for i, x in enumerate(list1) if x in list2]

def check_R_packages(required_packages, R_path):
	'''
	Determines whether or not R packages are properly installed
	'''
	installed_packages = os.popen(R_path + ''' -e "installed.packages()" | awk '{print $1}' | sort | uniq''').read().strip().split("\n")
	if(not set(required_packages) < set(installed_packages)):
		sys.exit("ERROR: cannot find the following R package: " + str(set(required_packages) - set(installed_packages)) + "\n" + 
			"Install it in your R console and then try rerunning proatac (but there may be other missing dependencies).")
		

def inferSampleVectors(input):
	'''
	Given input file / directory, creates 3 vectors of samplename, fastq1, fastq2
	'''
	def list_duplicates_of(seq,item):
		start_at = -1
		locs = []
		while True:
			try:
				loc = seq.index(item,start_at+1)
			except ValueError:
				break
			else:
				locs.append(loc)
				start_at = loc
		return locs

	samplenames = []
	fastq1 = []
	fastq2 = []

	# Import sample table if that's what is provided
	if(os.path.isfile(input)):
		sniff_range = 4096
		delimiters = ';\t, '
		sniffer = csv.Sniffer()

		with open(input, 'r') as infile:
			# Determine dialect
			dialect = sniffer.sniff(
				infile.read(sniff_range), delimiters=delimiters
			)
			infile.seek(0)

			# Sniff for header
			has_header = sniffer.has_header(infile.read(sniff_range))
			infile.seek(0)
			reader = csv.reader(infile, dialect)
			for row in reader:
				if(len(row) > 0):
					while '' in row:
						row.remove('')
					samplenames.append(row[0])
					fastq1.append(verify_file(row[1]))
					fastq2.append(verify_file(row[2]))
	
	# Otherwise figure out .fastq files from directory, do the merging, and infer sample names
	else:
		files = os.popen("ls " + input.rstrip("/") +"/*.fastq.gz").read().strip().split("\n")
		if(len(files) < 2):
			sys.exit("ERROR: input determined to be a directory but no paired .fastq.gz files found; QUITTING")
		files1 = [filename.replace("_R1", "_1").replace("_R2", "_1").replace("_2", "_1") for filename in files]
		samples = [os.path.basename(sample).split("_1")[0] for sample in files1]
		dups_in_source = partial(list_duplicates_of, samples)
		for c in set(samples):
			if(len(dups_in_source(c)) == 2):
				samplenames.append(c)
				fastq1.append(verify_file(files[dups_in_source(c)[0]]))
				fastq2.append(verify_file(files[dups_in_source(c)[1]]))

	return(samplenames, fastq1, fastq2)

def filterExistingSamples(samples, fastq1, fastq2, output):
	'''
	Looks in the output folder to see which samples have been 
	terminally processed already and will filter them from the three vectors
	of sample names, fastqs, etc. 
	'''
	# use findIdx here to infer which samples
	# already exist
	return(samples, fastq1, fastq2)

# https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python	
def available_cpu_count():
	'''
	Number of available virtual or physical CPUs on this system, i.e.
	user/real as output by time(1) when called with an optimally scaling
	userspace-only program
	'''
	# cpuset
	# cpuset may restrict the number of *available* processors
	try:
		m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
					  open('/proc/self/status').read())
		if m:
			res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
			if res > 0:
				return res
	except IOError:
		pass

	# Python 2.6+
	try:
		import multiprocessing
		return multiprocessing.cpu_count()
	except (ImportError, NotImplementedError):
		pass

	# http://code.google.com/p/psutil/
	try:
		import psutil
		return psutil.cpu_count()   # psutil.NUM_CPUS on old versions
	except (ImportError, AttributeError):
		pass

	# POSIX
	try:
		res = int(os.sysconf('SC_NPROCESSORS_ONLN'))

		if res > 0:
			return res
	except (AttributeError, ValueError):
		pass

	# Windows
	try:
		res = int(os.environ['NUMBER_OF_PROCESSORS'])

		if res > 0:
			return res
	except (KeyError, ValueError):
		pass

	# jython
	try:
		from java.lang import Runtime
		runtime = Runtime.getRuntime()
		res = runtime.availableProcessors()
		if res > 0:
			return res
	except ImportError:
		pass

	# BSD
	try:
		sysctl = subprocess.Popen(['sysctl', '-n', 'hw.ncpu'],
								  stdout=subprocess.PIPE)
		scStdout = sysctl.communicate()[0]
		res = int(scStdout)

		if res > 0:
			return res
	except (OSError, ValueError):
		pass

	# Linux
	try:
		res = open('/proc/cpuinfo').read().count('processor\t:')

		if res > 0:
			return res
	except IOError:
		pass

	# Solaris
	try:
		pseudoDevices = os.listdir('/devices/pseudo/')
		res = 0
		for pd in pseudoDevices:
			if re.match(r'^cpuid@[0-9]+$', pd):
				res += 1

		if res > 0:
			return res
	except OSError:
		pass

	# Other UNIXes (heuristic)
	try:
		try:
			dmesg = open('/var/run/dmesg.boot').read()
		except IOError:
			dmesgProcess = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE)
			dmesg = dmesgProcess.communicate()[0]

		res = 0
		while '\ncpu' + str(res) + ':' in dmesg:
			res += 1

		if res > 0:
			return res
	except OSError:
		pass

	raise Exception('Can not determine number of CPUs on this system')



import click
import os
import os.path
import sys
import shutil
import random
import string
import itertools
import time
import platform
from ruamel import yaml
from .bapHelp import *

def getBfiles(bedtools_genome, blacklist_file, reference_genome, script_dir, supported_genomes):

	'''
	Function that isn't actually a bapProject specific function.
	Used to collate the built-in genomes with the possibility that the
	user specified another genome.
	'''
	
	# Handle bedtools
	if(bedtools_genome == "" and reference_genome == ""):
		sys.exit("ERROR: bap needs either the bedtools genome or a correctly specified reference genome to get peaks from summit files; QUITTING")
	elif any(reference_genome == s for s in supported_genomes):
		bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + reference_genome + ".sizes"
	else:
		if(os.path.isfile(bedtools_genome)):
			bedtoolsGenomeFile = bedtools_genome
		else: 
			sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
	
	# Handle blacklist	
	if(blacklist_file == "" and reference_genome == ""):
		sys.exit("ERROR: bap needs either a blacklist bed file or a correctly specified reference genome to get peaks from summit files; QUITTING")
	elif any(reference_genome == s for s in supported_genomes):
		blacklistFile = script_dir + "/anno/blacklist/" + reference_genome + ".full.blacklist.bed"
	else:
		if(os.path.isfile(bedtools_genome)):
			blacklistFile = bedtools_genome
		else: 
			sys.exit("Could not find the blacklist file: %s" % bedtools_genome)
			
	return(bedtoolsGenomeFile, blacklistFile)

def mitoChr(reference_genome, mito_chromosome):
	if(mito_chromosome != "default"):
		return(mito_chromosome)
	else:
		if(reference_genome in ["hg19", "mm10", "hg38", "mm10"]):
			return("chrM")
		elif(reference_genome == "hg19_mm10_c"):
			return("humanM")
		else:
			return("hg19_chrM")

class bap1Project():
	def __init__(self, script_dir, supported_genomes, mode, input, output, name, ncores, reference_genome,
		cluster, jobs, peak_file,
		minimum_barcode_fragments, minimum_cell_fragments, barcode_whitelist,
		minimum_jaccard_index, nc_threshold, one_to_one, barcoded_tn5,
		extract_mito, keep_temp_files, mapq, 
		bedtools_genome, blacklist_file, tss_file, mito_chromosome, r_path, 
		drop_tag, bead_tag):
		
		#----------------------------------
		# Assign straightforward attributes
		#----------------------------------
		self.script_dir = script_dir
		self.mode = mode
		self.output = output
		self.cluster = cluster
		self.jobs = jobs
		self.mapq = mapq
		
		if(minimum_jaccard_index > 1):
			sys.exit("Cannot specify jaccard index > 1; user specified : %s" % str(minimum_jaccard_index))
		
		if(minimum_jaccard_index < 0):
			sys.exit("Cannot specify a nevative jaccard index; user specified : %s" % str(minimum_jaccard_index))
		
		
		# Handle potential whitelist
		if(barcode_whitelist == ""):
			barcode_whitelist = "none"
		else:
			if(os.path.isfile(barcode_whitelist)):
				pass
			else: 
				sys.exit("Could not find the bead whitelist file: %s" % barcode_whitelist)
		self.barcode_whitelist = barcode_whitelist
		
		self.minimum_barcode_fragments = minimum_barcode_fragments
		self.minimum_cell_fragments = minimum_cell_fragments
		self.nc_threshold = nc_threshold
		self.minimum_jaccard_index = minimum_jaccard_index
		self.one_to_one = one_to_one
		self.barcoded_tn5 = barcoded_tn5
		self.extract_mito = extract_mito
		self.drop_tag = drop_tag
		self.bead_tag = bead_tag
		
		# Figure out operating system just for funzies; not presently used
		self.os = "linux"
		if(platform.platform()[0:5]=="Darwi"):
			self.os = "mac"
		
		if(mode == "bam"):
			self.bamfile = input
			self.bwa = "NA"
			self.bwa_index = "bwa_index"
			self.name = name
			
			if(name == "default"):
				filename, file_extension = os.path.splitext(self.bamfile)
				self.name = os.path.basename(filename)

		#------------------------------------------
		# Verify R and all of its packages are here
		#------------------------------------------
		R = get_software_path('R', r_path)
		#check_R_packages(['Rsamtools', 'GenomicAlignments', 'GenomicRanges', 'dplyr'], R)
		self.R = R

		#------------------------
		# Handle reference genome
		#------------------------
		self.reference_genome = reference_genome
		if any(self.reference_genome == s for s in supported_genomes):
			click.echo(gettime() + "Found designated reference genome: %s" % self.reference_genome)
			
			self.tssFile = script_dir + "/anno/TSS/" + self.reference_genome + ".refGene.TSS.bed"
			self.blacklistFile = script_dir + "/anno/blacklist/" + self.reference_genome + ".full.blacklist.bed"
			self.bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + self.reference_genome + ".sizes"

		else: 
			click.echo(gettime() + "Could not identify this reference genome: %s" % self.reference_genome)
			click.echo(gettime() + "Attempting to infer necessary input files from user specification.")
			necessary = [bedtools_genome, blacklist_file, tss_file, macs2_genome_size, bs_genome]
			if '' in necessary:
				if reference_genome == '':
					sys.exit("ERROR: specify valid reference genome with --reference-genome flag; QUITTING")
				else:
					sys.exit("ERROR: non-supported reference genome specified so these five must be validly specified: --bedtools-genome, --blacklist-file, --tss-file; QUITTING")
		
		#------------------------------		
		# Make sure all files are valid
		#------------------------------	
		if(bedtools_genome != ""):
			if(os.path.isfile(bedtools_genome)):
				self.bedtoolsGenomeFile = bedtools_genome
			else: 
				sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
				
		if(blacklist_file != ""):
			if(os.path.isfile(blacklist_file)):
				self.blacklistFile = blacklist_file
			else: 
				sys.exit("Could not find the blacklist bed file: %s" % blacklist_file)
		
		if(tss_file != ""):	
			if(os.path.isfile(tss_file)):
				self.tssFile = tss_file
			else: 
				sys.exit("Could not find the transcription start sites file: %s" % tss_file)
		
		if(peak_file != ""):	
			if(os.path.isfile(peak_file)):
				self.peakFile = peak_file
			else: 
				sys.exit("Could not find the transcription start sites file: %s" % peak_file)
		else:
			self.peakFile = ""
		
		self.mitochr =  mitoChr(reference_genome, mito_chromosome)
		
		
	#--------------------------------------------------------------------------------
	# Define a method to dump the object as a .yaml/dictionary for use in other files
	#--------------------------------------------------------------------------------
	def __iter__(self):
		
		yield 'script_dir', self.script_dir
		yield 'mode', self.mode
		yield 'output', self.output
		yield 'bamfile', self.bamfile
		yield 'name', self.name

		yield 'cluster', self.cluster
		yield 'jobs', self.jobs
		yield 'peakFile', self.peakFile
		
		yield 'minimum_barcode_fragments', self.minimum_barcode_fragments
		yield 'minimum_cell_fragments', self.minimum_cell_fragments
		yield 'barcode_whitelist', self.barcode_whitelist

		yield 'minimum_jaccard_index', self.minimum_jaccard_index
		yield 'nc_threshold', self.nc_threshold
		yield 'one_to_one', self.one_to_one
		yield 'barcoded_tn5', self.barcoded_tn5
		
		yield 'extract_mito', self.extract_mito
		yield 'mapq', self.mapq
		
		yield 'tssFile', self.tssFile
		yield 'blacklistFile', self.blacklistFile
		yield 'bedtoolsGenomeFile', self.bedtoolsGenomeFile
		yield 'mitochr', self.mitochr
		yield 'R', self.R
		
		yield 'drop_tag', self.drop_tag
		yield 'bead_tag', self.bead_tag
		
	


import sys
import re
import pysam
import shutil
import os 
import gzip

from multiprocessing import Pool
from optparse import OptionParser
from collections import Counter
from contextlib import contextmanager

opts = OptionParser()
usage = "usage: %prog [options] [inputs] Script to process aligned .bam files to 1) split by chromosome and 2) report read ID with bead barcode"
opts = OptionParser(usage=usage)

opts.add_option("--input", "-i", help="Name of the .bam file to parse")
opts.add_option("--name", "-n", help="Name of the set of .bam files to collate")
opts.add_option("--output", "-o", help="Path to the output directory for these")

opts.add_option("--mapq", default = 30, help="Minimum mapq for a read to be kept")
opts.add_option("--barcode-tag", default = 'XB', help="Name of the first .bam file")
opts.add_option("--mito-chr", default = "chrM", help="Designation of mtDNA chromosome")
opts.add_option("--ncores", default = 4, help="Number of cores for parallel processing")
opts.add_option("--bedtools-reference-genome", default = "", help="Reference genome sizes from bedtools")

options, arguments = opts.parse_args()

bamname = options.input
name = options.name
out = options.output

minmapq = float(options.mapq)

barcodeTag = options.barcode_tag
mitochr = options.mito_chr

cpu = int(options.ncores)
bedtoolsGenomeFile = options.bedtools_reference_genome
def intersection(lst1, lst2):
	lst3 = [value for value in lst1 if value in lst2]
	return lst3

def listDictToCounter(lst):
	dct = Counter()
	for d in lst:
		for k, v in d.items():
			dct[k] += v 
	return(dct)

def getBarcode(intags):
	'''
	Parse out the barcode per-read
	'''
	for tg in intags:
		if(barcodeTag == tg[0]):
			return(tg[1])
	return("NA")


#---------------------------------------------------------
# Function for writing the read name and the bead barcode ID for later use
#---------------------------------------------------------
def writeBeadReadName(two):
	chrom = two[0]
	filename = two[1]
	idx = chrs.index(chrom)
	
	# Iterate through bam file
	bam = pysam.AlignmentFile(bamname,'rb')
	Itr = bam.fetch(str(chrom),multiple_iterators=True)
	with gzip.open(filename, 'wt') as out_write:
		for read in Itr:
			# only consider reads with sufficient mapping quality
			if(read.mapping_quality > minmapq):
				read_barcode = getBarcode(read.tags)
				read_name = read.query_name
				value = read_name + "\t" + read_barcode + "\n"
				out_write.write(value)
	bam.close()
	
	# Split into per-chromosome bam files
	new_bam = filename.replace(".read_bead.tsv.gz", ".raw.bam")
	pysam.view(bamname, chrom, '-b' ,'-o', new_bam, catch_stdout=False)
	pysam.index(new_bam)
	return(chrom)

# Handle the chromosomes
chrlens = {}
with open(bedtoolsGenomeFile) as f:
	for line in f:
		tok = line.split("\t")
		chrlens[tok[0]] = tok[1].strip()

chrlenpass = {x : chrlens[x] for x in chrlens }
chrs = list(chrlenpass.keys())
	
# Final loop to write out passing reads
read_barcode_file = [out + "/" + name + "." + chr + ".read_bead." +"tsv.gz" for chr in chrs]

pool = Pool(processes=cpu)
toy_out = pool.map(writeBeadReadName, zip(chrs, read_barcode_file))
pool.close()

# Make some routing files
bamchrfiles = [out + "/" + name + "." + chr + ".raw" +".bam" for chr in chrs if chr != mitochr]
bamchrrouter = open(out.replace("temp/filt_split", ".internal/samples") + "/" + name + ".chrbam.txt", "w") 
for v in bamchrfiles:
	bamchrrouter.write(v+"\n")
bamchrrouter.close() 

bamchrrouter2 = open(out.replace("temp/filt_split", ".internal/samples") + "/" + name + ".mitochrbam.txt", "w")
bamchrrouter2.write(out + "/" + name + "." + mitochr + ".raw" +".bam"+"\n")
bamchrrouter2.close() 







import os
import re
import regex
import sys
import gzip
from itertools import repeat

from fuzzywuzzy import fuzz
from fuzzywuzzy import process
from fuzzysearch import find_near_matches

def batch_iterator(iterator, batch_size):
	"""
	Returns lists of tuples of length batch_size.
	"""
	entry = True  # Make sure we loop once
	while entry:
		batch = []
		while len(batch) < batch_size:
			try:
				entry = iterator.__next__()
			except StopIteration:
				entry = None
			if entry is None:
				# End of file
				break
			batch.append(entry)
		if batch:
			yield batch


def chunk_writer_gzip(filename, what):
	'''
	Basic function to write a chunk of a fastq file
	to a gzipped file
	'''
	with gzip.open(filename, 'wt') as out_write:
				out_write.writelines(what)
	return(filename)			

def prove_barcode_simple(bc, valid_set):
	'''
	Function that takes a putative barcode and returns the nearest valid one
	'''
		
	if(bc in valid_set):
		return(bc)
	else:
		return("NA")

def prove_barcode(bc, valid_set, n_mismatch):
	'''
	Function that takes a putative barcode and returns the nearest valid one
	'''
		
	if(bc in valid_set):
		return(bc, "0")
	else:
		eo = process.extractOne(bc, valid_set)
		if(eo[1] >= (len(bc)-n_mismatch)/len(bc)*100): 
			return(eo[0], "1")
		else:
			return("N"*len(bc), "0")
			
def formatRead(title, sequence, quality):
	"""
	Takes three components of fastq file and stiches them together in a string
	"""
	return("@%s\n%s\n+\n%s\n" % (title, sequence, quality))



import click
import os
import os.path
import sys
import shutil
import random
import string
import itertools
import time
import platform
from ruamel import yaml
from pkg_resources import get_distribution
from .bapHelp import *

def getBfiles(bedtools_genome, blacklist_file, reference_genome, script_dir, supported_genomes):

	'''
	Function that isn't actually a bapProject specific function.
	Used to collate the built-in genomes with the possibility that the
	user specified another genome.
	'''
	
	# Handle bedtools
	if(bedtools_genome == "" and reference_genome == ""):
		sys.exit("ERROR: bap needs either the bedtools genome or a correctly specified reference genome to get peaks from summit files; QUITTING")
	elif any(reference_genome == s for s in supported_genomes):
		bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + reference_genome + ".sizes"
	else:
		if(os.path.isfile(bedtools_genome)):
			bedtoolsGenomeFile = bedtools_genome
		else: 
			sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
	
	# Handle blacklist	
	if(blacklist_file == "" and reference_genome == ""):
		sys.exit("ERROR: bap needs either a blacklist bed file or a correctly specified reference genome to get peaks from summit files; QUITTING")
	elif any(reference_genome == s for s in supported_genomes):
		blacklistFile = script_dir + "/anno/blacklist/" + reference_genome + ".full.blacklist.bed"
	else:
		if(os.path.isfile(bedtools_genome)):
			blacklistFile = bedtools_genome
		else: 
			sys.exit("Could not find the blacklist file: %s" % bedtools_genome)
			
	return(bedtoolsGenomeFile, blacklistFile)

def mitoChr(reference_genome, mito_chromosome):

	if(mito_chromosome != "default"):
		return(mito_chromosome)
	else:
		if(reference_genome in ["hg19", "mm10", "hg38", "mm10"]):
			return("chrM")
		elif(reference_genome in ["GRCh37", "GRCh38", "GRCm37", "GRCm38"]):
			return("MT")
		elif(reference_genome == "hg19_mm10_c"):
			return("humanM")
		else:
			return("hg19_chrM")

class bapFragProject():
	def __init__(self, script_dir, supported_genomes, input, output, name, ncores, reference_genome,
		cluster, jobs, barcode_translate,
		nc_threshold, keep_temp_files, mapq, 
		bedtools_genome, blacklist_file, tss_file, mito_chromosome,
		r_path, bedtools_path, samtools_path, bgzip_path, tabix_path,
		bead_tag):
		
		#----------------------------------
		# Assign straightforward attributes
		#----------------------------------
		self.bap_version = get_distribution('bap').version
		self.script_dir = script_dir
		self.bamfile = input
		self.name = name
		self.output = output
		self.cluster = cluster
		self.jobs = jobs
		self.mapq = mapq
		self.ncores = ncores
		self.nc_threshold = nc_threshold
		self.bead_tag = bead_tag
		# Figure out operating system just for funzies; not presently used
		self.os = "linux"
		if(platform.platform()[0:5]=="Darwi"):
			self.os = "mac"
		
		if(name == "default"):
			filename, file_extension = os.path.splitext(self.bamfile)
			self.name = os.path.basename(filename)

		#------------------------------------------
		# Verify R and all of its packages are here
		#------------------------------------------
		R = get_software_path('R', r_path)
		#check_R_packages(['Rsamtools', 'GenomicAlignments', 'GenomicRanges', 'dplyr'], R)
		self.R = R
		
		bedtools = get_software_path('bedtools', bedtools_path)
		self.bedtools = bedtools
		samtools = get_software_path('samtools', samtools_path)
		self.samtools = samtools
		bgzip = get_software_path('bgzip', bgzip_path)
		self.bgzip = bgzip
		tabix = get_software_path('tabix', tabix_path)
		self.tabix = tabix
		
		#------------------------
		# Handle reference genome
		#------------------------
		self.reference_genome = reference_genome
		if any(self.reference_genome == s for s in supported_genomes):
			click.echo(gettime() + "Found designated reference genome: %s" % self.reference_genome)
			
			self.tssFile = script_dir + "/anno/TSS/" + self.reference_genome + ".refGene.TSS.bed"
			self.blacklistFile = script_dir + "/anno/blacklist/" + self.reference_genome + ".full.blacklist.bed"
			self.bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + self.reference_genome + ".sizes"

		else: 
			click.echo(gettime() + "Could not identify this reference genome: %s" % self.reference_genome)
			click.echo(gettime() + "Attempting to infer necessary input files from user specification.")
			necessary = [bedtools_genome, blacklist_file, tss_file, macs2_genome_size, bs_genome]
			if '' in necessary:
				if reference_genome == '':
					sys.exit("ERROR: specify valid reference genome with --reference-genome flag; QUITTING")
				else:
					sys.exit("ERROR: non-supported reference genome specified so these five must be validly specified: --bedtools-genome, --blacklist-file, --tss-file; QUITTING")
		
		if(reference_genome in ["hg19-mm10", "hg19_mm10_c"]):
			self.speciesMix = "yes"
		else:
			self.speciesMix = "none"
		
		if(os.path.isfile(barcode_translate)):
			self.barcodeTranslateFile = barcode_translate
		else: 
			sys.exit("Could not find the barcodes translate file: %s" % barcode_translate)
		
		#------------------------------		
		# Make sure all files are valid
		#------------------------------	
		if(bedtools_genome != ""):
			if(os.path.isfile(bedtools_genome)):
				self.bedtoolsGenomeFile = bedtools_genome
			else: 
				sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
				
		if(blacklist_file != ""):
			if(os.path.isfile(blacklist_file)):
				self.blacklistFile = blacklist_file
			else: 
				sys.exit("Could not find the blacklist bed file: %s" % blacklist_file)
		
		if(tss_file != ""):	
			if(os.path.isfile(tss_file)):
				self.tssFile = tss_file
			else: 
				sys.exit("Could not find the transcription start sites file: %s" % tss_file)
		
		self.mitochr = mitoChr(reference_genome, mito_chromosome)
		
	#--------------------------------------------------------------------------------
	# Define a method to dump the object as a .yaml/dictionary for use in other files
	#--------------------------------------------------------------------------------
	def __iter__(self):
		
		yield 'bap_version', self.bap_version
		yield 'script_dir', self.script_dir
		yield 'output', self.output
		yield 'bamfile', self.bamfile
		yield 'name', self.name
		yield 'ncores', self.ncores
		
		yield 'cluster', self.cluster
		yield 'jobs', self.jobs
		
		yield 'barcodeTranslateFile', self.barcodeTranslateFile

		yield 'nc_threshold', self.nc_threshold
		yield 'mapq', self.mapq
		
		yield 'tssFile', self.tssFile
		yield 'blacklistFile', self.blacklistFile
		yield 'bedtoolsGenomeFile', self.bedtoolsGenomeFile
		yield 'mitochr', self.mitochr
		
		yield 'R', self.R
		yield 'samtools', self.samtools
		yield 'bedtools', self.bedtools
		yield 'bgzip', self.bgzip
		yield 'tabix', self.tabix
		
		yield 'bead_tag', self.bead_tag

class bapBulkFragProject():
	def __init__(self, script_dir, supported_genomes, input, output, name, ncores, reference_genome,
		cluster, jobs,
		keep_temp_files, mapq, 
		bedtools_genome, blacklist_file, tss_file, mito_chromosome,
		r_path, bedtools_path, samtools_path, bgzip_path, tabix_path):
		
		#----------------------------------
		# Assign straightforward attributes
		#----------------------------------
		self.bap_version = get_distribution('bap').version
		self.script_dir = script_dir
		self.bamfile = input
		self.name = name
		self.output = output
		self.cluster = cluster
		self.jobs = jobs
		self.mapq = mapq
		self.ncores = ncores

		# Figure out operating system just for funzies; not presently used
		self.os = "linux"
		if(platform.platform()[0:5]=="Darwi"):
			self.os = "mac"
		
		if(name == "default"):
			filename, file_extension = os.path.splitext(self.bamfile)
			self.name = os.path.basename(filename)

		#------------------------------------------
		# Verify R and all of its packages are here
		#------------------------------------------
		R = get_software_path('R', r_path)
		#check_R_packages(['Rsamtools', 'GenomicAlignments', 'GenomicRanges', 'dplyr'], R)
		self.R = R
		
		bedtools = get_software_path('bedtools', bedtools_path)
		self.bedtools = bedtools
		samtools = get_software_path('samtools', samtools_path)
		self.samtools = samtools
		bgzip = get_software_path('bgzip', bgzip_path)
		self.bgzip = bgzip
		tabix = get_software_path('tabix', tabix_path)
		self.tabix = tabix
		
		#------------------------
		# Handle reference genome
		#------------------------
		self.reference_genome = reference_genome
		if any(self.reference_genome == s for s in supported_genomes):
			click.echo(gettime() + "Found designated reference genome: %s" % self.reference_genome)
			
			self.tssFile = script_dir + "/anno/TSS/" + self.reference_genome + ".refGene.TSS.bed"
			self.blacklistFile = script_dir + "/anno/blacklist/" + self.reference_genome + ".full.blacklist.bed"
			self.bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + self.reference_genome + ".sizes"

		else: 
			click.echo(gettime() + "Could not identify this reference genome: %s" % self.reference_genome)
			click.echo(gettime() + "Attempting to infer necessary input files from user specification.")
			necessary = [bedtools_genome, blacklist_file, tss_file, macs2_genome_size, bs_genome]
			if '' in necessary:
				if reference_genome == '':
					sys.exit("ERROR: specify valid reference genome with --reference-genome flag; QUITTING")
				else:
					sys.exit("ERROR: non-supported reference genome specified so these five must be validly specified: --bedtools-genome, --blacklist-file, --tss-file; QUITTING")
		
		if(reference_genome in ["hg19-mm10", "hg19_mm10_c"]):
			self.speciesMix = "yes"
		else:
			self.speciesMix = "none"
		
		#------------------------------		
		# Make sure all files are valid
		#------------------------------	
		if(bedtools_genome != ""):
			if(os.path.isfile(bedtools_genome)):
				self.bedtoolsGenomeFile = bedtools_genome
			else: 
				sys.exit("Could not find the bedtools genome file: %s" % bedtools_genome)
				
		if(blacklist_file != ""):
			if(os.path.isfile(blacklist_file)):
				self.blacklistFile = blacklist_file
			else: 
				sys.exit("Could not find the blacklist bed file: %s" % blacklist_file)
		
		if(tss_file != ""):	
			if(os.path.isfile(tss_file)):
				self.tssFile = tss_file
			else: 
				sys.exit("Could not find the transcription start sites file: %s" % tss_file)
		
		self.mitochr = mitoChr(reference_genome, mito_chromosome)
		
	#--------------------------------------------------------------------------------
	# Define a method to dump the object as a .yaml/dictionary for use in other files
	#--------------------------------------------------------------------------------
	def __iter__(self):
		
		yield 'bap_version', self.bap_version
		yield 'script_dir', self.script_dir
		yield 'output', self.output
		yield 'bamfile', self.bamfile
		yield 'name', self.name
		yield 'ncores', self.ncores
		
		yield 'cluster', self.cluster
		yield 'jobs', self.jobs
		yield 'mapq', self.mapq
		
		yield 'tssFile', self.tssFile
		yield 'blacklistFile', self.blacklistFile
		yield 'bedtoolsGenomeFile', self.bedtoolsGenomeFile
		yield 'mitochr', self.mitochr
		
		yield 'R', self.R
		yield 'samtools', self.samtools
		yield 'bedtools', self.bedtools
		yield 'bgzip', self.bgzip
		yield 'tabix', self.tabix
		
	


"""
bap: Bead-based scATAC-seq data Processing
"""
from setuptools import find_packages, setup
from distutils.core import setup, Extension

dependencies = ['python-Levenshtein','biopython','fuzzysearch','fuzzywuzzy','click', 'pytest', 'snakemake', 'optparse-pretty', 'regex', 'pysam', 'ruamel.yaml']

setup(
    name='bap-atac',
    version='0.6.4',
    url='https://github.com/caleblareau/bap',
    license='MIT',
    author='Caleb Lareau',
    author_email='clareau@broadinstitute.org',
    description='Bead-based scATAC-seq data Processing.',
    long_description=__doc__,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    entry_points={
        'console_scripts': [
            'bap = bap.cli_old_dontEdit:main',
            'bap-barcode = bap.barcode.cli_barcode:main',
            'bap2 = bap.cli_bap2:main',
            'bap-frag = bap.cli_bap_frag:main',
            'bap-bulk-frag = bap.cli_bap_bulk_frag:main',
            'bap-reanno = bap.cli_reanno:main'
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
         'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



#!/usr/bin/env python

import sys
import re
import pysam
import gzip
from optparse import OptionParser

opts = OptionParser()
usage = "usage: %prog [options] [inputs] Software to process aligned bam files and annotate with a drop barcode"
opts = OptionParser(usage=usage)
opts.add_option("--input", help="Filename of the existing .bam file")
opts.add_option("--output", help="Filename of new .bam file to be generated")
opts.add_option("--bead-barcode", help="Read tag of the bead barcode")
opts.add_option("--drop-barcode", help="Read tag of the drop barcode (to be added)")
opts.add_option("--dict-file", help="Filepath of the dictionary to convert bead-barcode to drop-barcode")
opts.add_option("--hq-frags", help="Filepath of the dictionary to file of deduplicated fragments to retain")

options, arguments = opts.parse_args()

bamname = options.input
outname = options.output
bb = options.bead_barcode
db = options.drop_barcode
dictfile = options.dict_file
hqfrags = options.hq_frags

def getBarcode(intags):
	'''
	Parse out the bead barcode per-read
	'''
	for tg in intags:
		if(bb == tg[0]):
			return(tg[1])
	return("NA")

# Handle a dictionary of bead-barcode : drop-barcode pairs
d = {}
with open(dictfile) as f:
  for line in f:
    tok = line.split()
    d[tok[0]] = tok[1].strip()

# Import a set of read names that are in the HQ deduplicated set
keep_reads = set( (x.split("\t"))[4].strip() for x in open(hqfrags, "rt") )

# Iterate through bam
bam = pysam.AlignmentFile(bamname, "rb")
out = pysam.AlignmentFile(outname, "wb", template = bam)

try:
	for read in bam:
		bead_bc = getBarcode(read.tags)
		drop_bc = d.get(bead_bc, "NA")
	
		# Handle droplet barcodes that we want to consider writing out
		if(drop_bc != "NA"):
			rn = read.query_name
			
			# Look up corresponding read name
			if(rn in keep_reads):
				read.tags = read.tags + [(db, drop_bc)]
				out.write(read)

except OSError: # Truncated bam 
	print('Finished parsing bam')
	
bam.close()
out.close()
